#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import json
import os

from copy import deepcopy
from typing import TYPE_CHECKING, List, Union
from warnings import warn
from contextlib import redirect_stdout

from numpy import ndarray
from pandas import DataFrame

from ibm_watson_machine_learning.helpers.connections import (
    DataConnection, S3Location, FSLocation, AssetLocation, WSDAssetLocation, ContainerLocation, DatabaseLocation)
from ibm_watson_machine_learning.preprocessing import DataJoinPipeline
from ibm_watson_machine_learning.utils.autoai.enums import (
    RunStateTypes, PipelineTypes, TShirtSize, ClassificationAlgorithms, RegressionAlgorithms, DataConnectionTypes,
    ForecastingPipelineTypes, PredictionType, TimeseriesAnomalyPredictionPipelineTypes)
from ibm_watson_machine_learning.utils.autoai.errors import (
    FitNotCompleted, MissingDataPreprocessingStep, WrongDataJoinGraphNodeName, DataSourceSizeNotSupported,
    TrainingDataSourceIsNotFile, NoneDataConnection, PipelineNotLoaded, OBMForNFSIsNotSupported,
    ForecastingUnsupportedOperation, LibraryNotCompatible, InvalidDataAsset, TestDataNotPresent,
    FutureExogenousFeaturesNotSupported)
from ibm_watson_machine_learning.utils.autoai.utils import try_import_lale, all_logging_disabled
from ibm_watson_machine_learning.utils.autoai.wsd_ui import (
    save_computed_pipelines_for_ui, save_experiment_for_ui, save_metadata_for_ui)
from ibm_watson_machine_learning.utils.autoai.connection import validate_source_data_connections, \
    validate_results_data_connection
from ibm_watson_machine_learning.utils import DisableWarningsLogger, WMLClientError
from ibm_watson_machine_learning.messages.messages import Messages
from .base_auto_pipelines import BaseAutoPipelines

if TYPE_CHECKING:
    from ibm_watson_machine_learning.experiment.autoai.engines import WMLEngine
    from ibm_watson_machine_learning.utils.autoai.enums import Metrics, PredictionType, Transformers
    from ibm_watson_machine_learning.preprocessing import DataJoinGraph
    from sklearn.pipeline import Pipeline

__all__ = [
    "RemoteAutoPipelines"
]


class RemoteAutoPipelines(BaseAutoPipelines):
    """RemoteAutoPipelines class for pipeline operation automation on WML.

    :param name: name for the AutoPipelines
    :type name: str

    :param prediction_type: type of the prediction
    :type prediction_type: PredictionType

    :param prediction_column: name of the target/label column
    :type prediction_column: str

    :param scoring: type of the metric to optimize with
    :type scoring: Metrics

    :param engine: engine for remote work on WML
    :type engine: WMLEngine

    :param desc: description
    :type desc: str, optional

    :param holdout_size: percentage of the entire dataset to leave as a holdout, default 0.1
    :type holdout_size: float, optional

    :param max_num_daub_ensembles: maximum number (top-K ranked by DAUB model selection) of the selected algorithm,
        or estimator types, for example `LGBMClassifierEstimator`, `XGBoostClassifierEstimator`, or
        `LogisticRegressionEstimator` to use in pipeline composition, the default is `None` that means
        the true default value will be determined by the internal different algorithms, where only
        the highest ranked by model selection algorithm type is used
    :type max_num_daub_ensembles: int, optional

    :param train_sample_rows_test_size: training data sampling percentage
    :type train_sample_rows_test_size: float, optional

    :param include_only_estimators: list of estimators to include in computation process
    :type include_only_estimators: list[ClassificationAlgorithms or RegressionAlgorithms], optional

    :param cognito_transform_names: list of transformers to include in the feature enginnering computation process,
        see: AutoAI.Transformers
    :type cognito_transform_names: list[Transformers], optional

    :param csv_separator: the separator, or list of separators to try for separating columns in a CSV file,
        not used if the file_name is not a CSV file, default is ','
    :type csv_separator: list[str] or str, optional

    :param excel_sheet: name of the excel sheet to use, only use when xlsx file is an input,
        support for number of the sheet is deprecated, by default first sheet is used
    :type excel_sheet: str, optional

    :param encoding: encoding type for CSV training file
    :type encoding: str, optional

    :param positive_label: the positive class to report when binary classification, when multiclass or regression,
        this will be ignored
    :type positive_label: str, optional

    :param t_shirt_size: the size of the remote AutoAI POD instance (computing resources),
        only applicable to a remote scenario
    :type t_shirt_size: TShirtSize, optional

    :param data_join_graph: a graph object with definition of join structure for multiple input data sources,
        data preprocess step for multiple files
    :type data_join_graph: DataJoinGraph, optional

    """

    def __init__(self,
                 name: str,
                 prediction_type: 'PredictionType',
                 prediction_column: str,
                 prediction_columns: List[str],
                 timestamp_column_name: str,
                 engine: 'WMLEngine',
                 scoring: 'Metrics' = None,
                 desc: str = None,
                 holdout_size: float = None,
                 max_num_daub_ensembles: int = None,
                 t_shirt_size: 'TShirtSize' = TShirtSize.M,
                 train_sample_rows_test_size: float = None,
                 include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']] = None,
                 include_batched_ensemble_estimators: List[Union['BatchedClassificationAlgorithms',
                                                                 'BatchedRegressionAlgorithms']] = None,
                 backtest_num: int = None,
                 lookback_window: int = None,
                 forecast_window: int = None,
                 backtest_gap_length: int = None,
                 cognito_transform_names: List['Transformers'] = None,
                 data_join_graph: 'DataJoinGraph' = None,
                 csv_separator: Union[List[str], str] = ',',
                 excel_sheet: Union[str, int] = None,
                 encoding: str = 'utf-8',
                 positive_label: str = None,
                 data_join_only: bool = False,
                 drop_duplicates: bool = True,
                 outliers_columns: list = None,
                 text_processing: bool = True,
                 word2vec_feature_number: int = None,
                 daub_give_priority_to_runtime: float = None,
                 notebooks=False,
                 autoai_pod_version=None,
                 obm_pod_version=None,
                 text_columns_names=None,
                 n_parallel_data_connections=None,
                 test_data_csv_separator: Union[List[str], str] = ',',
                 test_data_excel_sheet: Union[str, int] = None,
                 test_data_encoding: str = 'utf-8',
                 sampling_type=None,
                 sample_size_limit=None,
                 sample_rows_limit=None,
                 sample_percentage_limit=None,
                 number_of_batch_rows=None,
                 categorical_imputation_strategy=None,
                 numerical_imputation_strategy=None,
                 numerical_imputation_value=None,
                 imputation_threshold=None,
                 fairness_info: dict = None,
                 retrain_on_holdout: bool = True,
                 feature_columns: List[str] = None,
                 pipeline_types: List[Union['ForecastingPipelineTypes', 'TimeseriesAnomalyPredictionPipelineTypes']] = None,
                 supporting_features_at_forecast: bool = None,
                 categorical_columns: list = None,
                 numerical_columns: list = None,
                 confidence_level: float = None,
                 incremental_learning: bool = None,
                 early_stop_enabled: bool = None,
                 early_stop_window_size: int = None,
                 **kwargs):

        if data_join_graph is not None:
            data_join_graph.fairness_info = fairness_info

        # Deprecation of excel_sheet as number:
        if isinstance(excel_sheet, int) or isinstance(test_data_excel_sheet, int):
            warn(
                message="Support for excel sheet as number of the sheet (int) is deprecated! Please set excel sheet with name of the sheet.")

        self.params = {
            'name': name,
            'desc': desc if desc else '',
            'prediction_type': prediction_type if prediction_type != 'timeseries' else 'forecasting',
            'prediction_column': prediction_column,
            'prediction_columns': prediction_columns,
            'timestamp_column_name': timestamp_column_name,
            'scoring': scoring,
            'holdout_size': holdout_size,
            'max_num_daub_ensembles': max_num_daub_ensembles,
            't_shirt_size': t_shirt_size,
            'train_sample_rows_test_size': train_sample_rows_test_size,
            'include_only_estimators': include_only_estimators,
            'include_batched_ensemble_estimators': include_batched_ensemble_estimators,
            'backtest_num': backtest_num,
            'lookback_window': lookback_window,
            'forecast_window': forecast_window,
            'backtest_gap_length': backtest_gap_length,
            'cognito_transform_names': cognito_transform_names,
            'data_join_graph': data_join_graph or False,
            'csv_separator': csv_separator,
            'excel_sheet': excel_sheet,
            'encoding': encoding,
            'positive_label': positive_label,
            'data_join_only': data_join_only,
            'drop_duplicates': drop_duplicates,
            'outliers_columns': outliers_columns,
            'notebooks': notebooks,
            'autoai_pod_version': autoai_pod_version,
            'obm_pod_version': obm_pod_version,
            'text_processing': text_processing,
            'word2vec_feature_number': word2vec_feature_number,
            'daub_give_priority_to_runtime': daub_give_priority_to_runtime,
            'text_columns_names': text_columns_names,
            'sampling_type': sampling_type,
            'sample_size_limit': sample_size_limit,
            'sample_rows_limit': sample_rows_limit,
            'sample_percentage_limit': sample_percentage_limit,
            'number_of_batch_rows': number_of_batch_rows,
            'n_parallel_data_connections': n_parallel_data_connections,
            'test_data_csv_separator': test_data_csv_separator,
            'test_data_excel_sheet': test_data_excel_sheet,
            'test_data_encoding': test_data_encoding,
            'categorical_imputation_strategy': categorical_imputation_strategy,
            'numerical_imputation_strategy': numerical_imputation_strategy,
            'numerical_imputation_value': numerical_imputation_value,
            'imputation_threshold': imputation_threshold,
            'retrain_on_holdout': retrain_on_holdout,
            'feature_columns': feature_columns,
            'pipeline_types': pipeline_types,
            'supporting_features_at_forecast': supporting_features_at_forecast,
            'numerical_columns': numerical_columns,
            'categorical_columns': categorical_columns,
            'confidence_level': confidence_level,
            'incremental_learning': incremental_learning,
            'early_stop_enabled': early_stop_enabled,
            'early_stop_window_size': early_stop_window_size
        }

        if fairness_info:
            self.params['fairness_info'] = fairness_info

        self._engine: 'WMLEngine' = engine
        self._engine.initiate_remote_resources(params=self.params, **kwargs)
        self.best_pipeline = None
        self._workspace = None

    def _get_engine(self) -> 'WMLEngine':
        """Return WMLEngine for development purposes."""
        return self._engine

    ####################################################
    #   WML Pipeline Part / Parameters for AUtoAI POD  #
    ####################################################
    def get_params(self) -> dict:
        """Get configuration parameters of AutoPipelines.

        :return: AutoPipelines parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.get_params()

            # Result:
            # {
            #     'name': 'test name',
            #     'desc': 'test description',
            #     'prediction_type': 'classification',
            #     'prediction_column': 'y',
            #     'scoring': 'roc_auc',
            #     'holdout_size': 0.1,
            #     'max_num_daub_ensembles': 1,
            #     't_shirt_size': 'm',
            #     'train_sample_rows_test_size': 0.8,
            #    'include_only_estimators': ["ExtraTreesClassifierEstimator",
            #                                "GradientBoostingClassifierEstimator",
            #                                "LGBMClassifierEstimator",
            #                                "LogisticRegressionEstimator",
            #                                "RandomForestClassifierEstimator",
            #                                "XGBClassifierEstimator"]
            # }
        """
        _params = self._engine.get_params().copy()
        del _params['autoai_pod_version']
        del _params['obm_pod_version']
        del _params['notebooks']
        del _params['data_join_only']

        return _params

    ###########################################################
    #   WML Training Part / Parameters for AUtoAI Experiment  #
    ###########################################################
    def fit(self,
            train_data: 'DataFrame' = None,
            *,
            training_data_reference: List['DataConnection'] = None,
            training_results_reference: 'DataConnection' = None,
            background_mode=False,
            test_data_references: List['DataConnection'] = None,
            training_data_references: List['DataConnection'] = None) -> dict:
        """Run a training process on WML of autoai on top of the training data referenced by DataConnection.

        :param training_data_reference: data storage connection details to inform where training data is stored,
            deprecated parameter, use `training_data_references` instead

        :param training_data_references: data storage connection details to inform where training data is stored,
            new version of `training_data_reference`
        :type training_data_references: list[DataConnection]

        :param training_results_reference: data storage connection details to store pipeline training results,
            not applicable on CP4D
        :type training_results_reference: DataConnection, optional

        :param background_mode: indicator if fit() method will run in background (async) or (sync)
        :type background_mode: bool, optional

        :param test_data_references: data storage connection details to inform where test / holdout data is stored
        :type test_data_references: list[DataConnection], optional

        :return: run details
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            from ibm_watson_machine_learning.helpers import DataConnection, S3Location

            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.fit(
                training_data_connection=[DataConnection(
                    connection_asset_id=connection_id,
                    location=S3Location(
                        bucket='automl',
                        path='german_credit_data_biased_training.csv')
                    )
                )],
                DataConnection(
                    connection_asset_id=connection_id,
                    location=S3Location(
                        bucket='automl',
                        path='')
                    )
                ),
                background_mode=False)
        """
        if self.params.get('data_join_graph'):
            if self._workspace.wml_client.ICP_46 or self._workspace.wml_client.ICP_47:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cpd"))
            elif self._workspace.wml_client.ICP:
                print(Messages.get_message(message_id="obm_deprecation_message_cpd"))
            else:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cloud"))

        if training_data_references is not None:
            training_data_reference = training_data_references

        if training_data_reference is None or not training_data_reference:
            raise NoneDataConnection('training_data_references')

        for conn in training_data_reference:
            if self._workspace.wml_client.project_type == 'local_git_storage':
                conn.location.userfs = 'true'
            conn.set_client(self._workspace.wml_client)
            # TODO: remove S3 implementation
            if conn.type == DataConnectionTypes.S3:
                conn._validate_cos_resource()

        training_data_reference = [new_conn for conn in training_data_reference for new_conn in
                                   conn._subdivide_connection()]

        if isinstance(test_data_references, list):
            for conn in test_data_references:
                # Update test data ref with client object, experiment parameters
                if isinstance(conn, DataConnection):
                    if self._workspace.wml_client.project_type == 'local_git_storage':
                        conn.location.userfs = 'true'
                    conn.set_client(self._workspace.wml_client)
                    # TODO: remove S3 implementation
                    if conn.type == DataConnectionTypes.S3:
                        conn._validate_cos_resource()

                    conn.auto_pipeline_params = self._engine._auto_pipelines_parameters

        # note: check if DataJoinGraph node names are correct and equivalent
        # to training DataConnections IDs/data_join_node_names
        if self.params['data_join_graph']:
            # note: If the location does not have path attribute it is either S3Location or DatabaseLocation
            if any(filter(lambda x: x.type == DataConnectionTypes.CA and hasattr(x.location, "path"),
                          training_data_reference)):
                raise OBMForNFSIsNotSupported()
            # --- end note

            data_join_graph = self.params['data_join_graph']
            data_connection_ids = [connection.id for connection in training_data_reference]
            for node_name in [node.table.name for node in data_join_graph.nodes]:
                if node_name not in data_connection_ids:
                    raise WrongDataJoinGraphNodeName(
                        node_name,
                        reason=f"Please make sure that each particular node name in data_join_graph is the same as "
                               f"\"data_join_node_name\" parameter in particular equivalent training DataConnection. "
                               f"The default names are taken as a DataConnection.location.path.")
        # --- end note

        # note: update each training data connection with pipeline parameters for holdout split recreation
        for data_connection in training_data_reference:
            data_connection.auto_pipeline_params = self._engine._auto_pipelines_parameters

        if isinstance(train_data, DataFrame):
            training_data_reference[0].write(data=train_data,
                                             remote_name=training_data_reference[0].location.path)
        elif train_data is None:
            pass

        else:
            raise TypeError("train_data should be of type pandas.DataFrame")

        # self._validate_training_data_size(training_data_reference)

        training_data_reference = validate_source_data_connections(training_data_reference, workspace=self._workspace,
                                                                   deployment=False)

        # note: for FSLocation we are creating asset and changing location to AssetLocation
        # so href is not set properly, setter on wml_client will resolve that issue
        for conn in training_data_reference:
            conn.set_client(self._workspace.wml_client)
        # --- end note

        training_results_reference = self.determine_result_reference(training_results_reference,
                                                                     training_data_reference,
                                                                     "default_autoai_out")

        if self.params.get('data_join_graph') and test_data_references: # only OBM scenario case
            test_output_data = self.determine_result_reference(None,
                                                               test_data_references,
                                                               "default_autoai_test_out")
        else:
            test_output_data = None

        run_params = self._engine.fit(training_data_reference=training_data_reference,
                                      training_results_reference=training_results_reference,
                                      background_mode=background_mode,
                                      test_data_references=test_data_references,
                                      test_output_data=test_output_data)

        for conn in training_data_reference:
            metrics = run_params['entity']['status'].get('metrics', [])
            if metrics and metrics[-1]['context'].get('fairness'):
                conn.auto_pipeline_params['fairness_info'] = metrics[-1]['context']['fairness'].get('info')

        if isinstance(training_data_reference[0].location, WSDAssetLocation) or (
                isinstance(training_data_reference[0].location, AssetLocation
                           ) and training_data_reference[0].location._wsd):
            try:
                wml_pipeline_details = self._workspace.wml_client.pipelines.get_details(
                    run_params['entity']['pipeline']['id'])
                save_experiment_for_ui(wml_pipeline_details,
                                       run_params,
                                       training_data_reference[0].location._local_asset_path,
                                       training_data_reference[0].location._asset_id,
                                       training_data_reference[0].location._asset_name)
                save_computed_pipelines_for_ui(self._workspace.wml_client,
                                               self._engine._current_run_id)
                save_metadata_for_ui(wml_pipeline_details, run_params)

            except Exception as e:
                print(f"Cannot save experiment locally. It will not be visible in the WSD UI. Error: {e}")
                warn(f"Cannot save experiment locally. It will not be visible in the WSD UI. Error: {e}")

        return run_params

    def determine_result_reference(self, results_reference, data_references, result_path):
        # note: if user did not provide results storage information, use default ones
        if results_reference is None:
            if isinstance(data_references[0].location, S3Location) and not self._workspace.wml_client.ICP:
                results_reference = DataConnection(
                    connection=data_references[0].connection,
                    location=S3Location(bucket=data_references[0].location.bucket,
                                        path=".")
                )

            elif isinstance(data_references[0].location, AssetLocation) and not self._workspace.wml_client.ICP:
                connection_id = data_references[0].location._get_connection_id(self._workspace.wml_client)

                if connection_id is not None:
                    results_reference = DataConnection(
                        connection_asset_id=connection_id,
                        location=S3Location(
                            bucket=data_references[0].location._get_bucket(self._workspace.wml_client),
                            path=result_path)
                    )

                else:  # set container output location when default DAta Asset is as a train ref
                    results_reference = DataConnection(
                        location=ContainerLocation(path=result_path))

            elif isinstance(data_references[0].location,
                            ContainerLocation) and not self._workspace.wml_client.ICP:
                results_reference = DataConnection(location=ContainerLocation(path=result_path))

            elif isinstance(data_references[0].location,
                            DatabaseLocation) and not self._workspace.wml_client.ICP:
                results_reference = DataConnection(location=ContainerLocation(path=result_path))

            else:
                location = FSLocation()
                if self._workspace.WMLS:
                    location.path = location.path.format(option='spaces',
                                                         id=self._engine._wml_client.default_space_id)
                else:
                    if self._workspace.wml_client.default_project_id is None:
                        location.path = location.path.format(option='spaces',
                                                             id=self._engine._wml_client.default_space_id)

                    else:
                        location.path = location.path.format(option='projects',
                                                             id=self._engine._wml_client.default_project_id)
                results_reference = DataConnection(
                    connection=None,
                    location=location
                )
        # -- end note
        if isinstance(results_reference.location, AssetLocation):
            if results_reference.location._get_connection_id(self._workspace.wml_client) is None:
                raise InvalidDataAsset(
                    reason="Please specify Data Asset pointing to connection e.g. COS as an output.")

        # note: results can be stored only on FS or COS
        if not isinstance(results_reference.location,
                          (S3Location, FSLocation, AssetLocation, ContainerLocation)):
            raise TypeError('Unsupported results location type. Results referance can be stored'
                            ' only on S3Location or FSLocation or AssetLocation.')
        # -- end

        # note: only if we are going with OBM + KB scenario, add ID to results DataConnection
        if self.params.get('data_join_graph'):
            results_reference.id = 'outputData'
        # --- end note

        return results_reference

    #####################
    #   Run operations  #
    #####################
    def get_run_status(self) -> str:
        """Check status/state of initialized AutoPipelines run if ran in background mode.

        :return: run status details
        :rtype: dict

        **Example**

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.get_run_status()

            # Result:
            # 'completed'
        """
        return self._engine.get_run_status()

    def get_run_details(self, include_metrics: bool = False, _internal=False) -> dict:
        """Get fit/run details.

        :param include_metrics: indicates to include metrics in the training details output
        :type include_metrics: bool, optional

        :return: AutoPipelineOptimizer fit/run details
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.get_run_details()
        """
        if not _internal and self.params.get('data_join_graph'):
            if self._workspace.wml_client.ICP_46 or self._workspace.wml_client.ICP_47:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cpd"))
            elif self._workspace.wml_client.ICP:
                print(Messages.get_message(message_id="obm_deprecation_message_cpd"))
            else:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cloud"))

        return self._engine.get_run_details(include_metrics=include_metrics)

    def cancel_run(self) -> None:
        """Cancels an AutoAI run."""
        self._engine.cancel_run()

    #################################
    #   Pipeline models operations  #
    #################################
    def summary(self, scoring: str = None, sort_by_holdout_score: bool = True) -> 'DataFrame':
        """Print AutoPipelineOptimizer Pipelines details (autoai trained pipelines).

        :param scoring: scoring metric which user wants to use to sort pipelines by,
            when not provided use optimized one
        :type scoring: string, optional

        :param sort_by_holdout_score: indicates if we want to sort pipelines by holdout metric or by training one,
            by default use holdout metric
        :type sort_by_holdout_score: bool, optional

        :return: computed pipelines and ML metrics
        :rtype: pandas.DataFrame

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.summary()

            # Result:
            #                training_normalized_gini_coefficient  ...  training_f1
            # Pipeline Name                                        ...
            # Pipeline_3                                 0.359173  ...     0.449197
            # Pipeline_4                                 0.359173  ...     0.449197
            # Pipeline_1                                 0.358124  ...     0.449057
            # Pipeline_2                                 0.358124  ...     0.449057
        """
        return self._engine.summary(scoring, sort_by_holdout_score)

    def get_pipeline_details(self, pipeline_name: str = None) -> dict:
        """Fetch specific pipeline details, eg. steps etc.

        :param pipeline_name: pipeline name eg. Pipeline_1, if not specified, best pipeline parameters will be fetched
        :type pipeline_name: str, optional

        :return: pipeline parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            remote_optimizer.get_pipeline_details()
            remote_optimizer.get_pipeline_details(pipeline_name='Pipeline_4')

            # Result:
            # {
            #     'composition_steps': ['TrainingDataset_full_4521_16', 'Split_TrainingHoldout',
            #                           'TrainingDataset_full_4068_16', 'Preprocessor_default', 'DAUB'],
            #     'pipeline_nodes': ['PreprocessingTransformer', 'GradientBoostingClassifierEstimator']
            # }
        """
        return self._engine.get_pipeline_details(pipeline_name=pipeline_name)

    def get_pipeline(self,
                     pipeline_name: str = None,
                     astype: 'PipelineTypes' = PipelineTypes.LALE,
                     persist: 'bool' = False) -> Union['Pipeline', 'TrainablePipeline']:
        """Download specified pipeline from WML.

        :param pipeline_name: pipeline name, if you want to see the pipelines names, please use summary() method,
            if this parameter is None, the best pipeline will be fetched
        :type pipeline_name: str, optional

        :param astype: type of returned pipeline model, if not specified, lale type is chosen
        :type astype: PipelineTypes, optional

        :param persist: indicates if selected pipeline should be stored locally
        :type persist: bool, optional

        :return: Scikit-Learn pipeline
        :rtype: Pipeline or TrainablePipeline

        See also RemoteAutoPipelines.summary().

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            pipeline_1 = remote_optimizer.get_pipeline(pipeline_name='Pipeline_1')
            pipeline_2 = remote_optimizer.get_pipeline(pipeline_name='Pipeline_1', astype=AutoAI.PipelineTypes.LALE)
            pipeline_3 = remote_optimizer.get_pipeline(pipeline_name='Pipeline_1', astype=AutoAI.PipelineTypes.SKLEARN)
            type(pipeline_3)
            # <class 'sklearn.pipeline.Pipeline'>
            pipeline_4 = remote_optimizer.get_pipeline(pipeline_name='Pipeline_1', persist=True)
            # Selected pipeline stored under: "absolute_local_path_to_model/model.pickle"

        """
        try:
            if pipeline_name is None:
                pipeline_model, check_lale = self._engine.get_best_pipeline(persist=persist)

            else:
                pipeline_model, check_lale = self._engine.get_pipeline(pipeline_name=pipeline_name, persist=persist)

        except ForecastingUnsupportedOperation as e:
            raise e

        except LibraryNotCompatible as e:
            raise e

        except Exception as e:
            raise PipelineNotLoaded(pipeline_name if pipeline_name is not None else 'best pipeline',
                                    reason=f"Pipeline with such a name probably does not exist. "
                                           f"Please make sure you specify correct pipeline name. Error: {e}")

        if astype == PipelineTypes.SKLEARN:
            return pipeline_model

        elif astype == PipelineTypes.LALE:
            if check_lale:
                try_import_lale()
            from lale.helpers import import_from_sklearn_pipeline
            with all_logging_disabled():
                # note: join preprocessing step to final pipeline (enables wider visualization and pretty print)
                try:
                    import numpy as np
                    preprocessing_object = self.get_preprocessing_pipeline()
                    preprocessing_pipeline = preprocessing_object.lale_pipeline

                    # note: fake fit for preprocessing pipeline to be able to predict
                    preprocessing_pipeline = preprocessing_pipeline.fit(np.array([[1, 2, 3], [2, 3, 4]]),
                                                                        np.array([1, 2]))
                    return preprocessing_pipeline >> import_from_sklearn_pipeline(pipeline_model)

                except MissingDataPreprocessingStep:
                    return import_from_sklearn_pipeline(pipeline_model)

                except BaseException as e:
                    message = f"Cannot load preprocessing step. Error: {e}"
                    print(message)
                    warn(message)
                    return import_from_sklearn_pipeline(pipeline_model)
                # --- end note

        else:
            raise ValueError('Incorrect value of \'astype\'. '
                             'Should be either PipelineTypes.SKLEARN or PipelineTypes.LALE')

    def get_pipeline_notebook(self,
                              pipeline_name: str = None,
                              filename: str = None,
                              insert_to_cell: bool = False) -> str:
        """Download specified pipeline notebook from WML.

        :param pipeline_name: pipeline name, if you want to see the pipelines names, please use summary() method,
            if this parameter is None, the best pipeline will be fetched
        :type pipeline_name: str, optional

        :param filename: filename under which the pipeline notebook will be saved
        :type filename: str, optional

        :param insert_to_cell: if run in jupyter notebook, if set inserts the notebook into cell below
        :type insert_to_cell: bool, optional

        :return: path to saved pipeline notebook
        :rtype: str

        See also RemoteAutoPipelines.summary().

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)
            remote_optimizer = experiment.optimizer(...)

            pipeline_notebook_path = remote_optimizer.get_pipeline_notebook(pipeline_name='Pipeline_1')

        """
        if pipeline_name is None:
            pipeline_name = self._engine.summary().index[0]

        path = self._engine.get_pipeline_notebook(pipeline_name, filename=filename)

        if insert_to_cell:
            with open(path, 'r') as f:
                content = json.loads(f.read())

            def translate_cell_to_str(cell):
                if cell['cell_type'] == 'code':
                    return ''.join(cell['source'])
                elif cell['cell_type'] == 'markdown':
                    return '\n'.join(["# " + l for l in ''.join(cell['source']).split('\n')])
                else:
                    return ''

            result = '\n\n'.join([translate_cell_to_str(c) for c in content['cells']])

            import sys

            def get_optimizer_var_name():
                for name, module in sys.modules.items():
                    for varname, obj in module.__dict__.items():
                        if obj is self:
                            return varname

            if '\'PUT_YOUR_APIKEY_HERE\'' in result:
                try:
                    optimizer_var_name = get_optimizer_var_name()
                    result = result.replace('\'PUT_YOUR_APIKEY_HERE\'', f'{optimizer_var_name}._workspace.wml_credentials[\'apikey\']')

                except:
                    pass

            import IPython.core
            ipython = IPython.core.getipython.get_ipython()
            comment = "# generated by get_pipeline_notebook(insert_to_cell=True) from previous cell\n\n"
            ipython.set_next_input(comment + result, replace=False)

        return path

    # note: predict on top of the best computed pipeline, best pipeline is downloaded for the first time
    def predict(self, X: Union['DataFrame', 'ndarray']=None, observations: Union['DataFrame', 'ndarray']=None,
                supporting_features: Union['DataFrame', 'ndarray']=None) -> 'ndarray':
        """Predict method called on top of the best fetched pipeline.

        :param X: test data for prediction
        :type X: numpy.ndarray or pandas.DataFrame

        :param observations: new observations of forecasting data that were used to train AutoAI model,
            supported only for forecasting pipelines
        :type observations: numpy.ndarray or pandas.DataFrame

        :param supporting_features: future values of exogenous features, supported only for forecasting pipelines
        :type supporting_features: numpy.ndarray or pandas.DataFrame

        :return: model predictions
        :rtype: numpy.ndarray
        """
        is_forecasting = self.params.get('prediction_type') in (PredictionType.FORECASTING, 'timeseries',
                                                                PredictionType.TIMESERIES_ANOMALY_PREDICTION)

        if not is_forecasting and X is None:
            raise ValueError("X parameter is required in predict method.")
        if is_forecasting:
            if X is not None and observations is not None:
                raise ValueError("For forecasting the  parameter X should not be used. Please use observations instead of X.")
            else:
                X = observations if observations is not None else X

        if self.best_pipeline is None:
            # note: automatically download the best computed pipeline
            if self.get_run_status() == RunStateTypes.COMPLETED:
                self.best_pipeline, _ = self._engine.get_best_pipeline()
            else:
                raise FitNotCompleted(self._engine._current_run_id,
                                      reason="Please check the run status with run_status() method.")
            # --- end note

        additional_params = {}
        if supporting_features is not None:
            if is_forecasting:
                additional_params.update(
                    {'supporting_features': supporting_features if isinstance(supporting_features,
                                                                           ndarray) else supporting_features.values})
            else:
                raise FutureExogenousFeaturesNotSupported(self.params.get('prediction_type'))

        if isinstance(X, DataFrame) or isinstance(X, ndarray):
            return self.best_pipeline.predict(X if isinstance(X, ndarray) else X.values, **additional_params)
        elif X is None and is_forecasting:
            return self.best_pipeline.predict()
        else:
            raise TypeError("X should be either of type pandas.DataFrame or numpy.ndarray")

    # --- end note

    def get_data_connections(self) -> List['DataConnection']:
        """Create DataConnection objects for further user usage
            (eg. to handle data storage connection or to recreate autoai holdout split).

        :return: list of DataConnection with populated optimizer parameters
        :rtype: list[DataConnection]
        """
        user_holdout_exists = True
        try:
            with DisableWarningsLogger():
                with open(os.devnull, "w") as f:
                    with redirect_stdout(f):
                        self.get_test_data_connections()

        except TestDataNotPresent:
            user_holdout_exists = False

        optimizer_parameters = self.get_params()
        training_data_references = self.get_run_details(_internal=True)['entity']['training_data_references']

        data_connections = [
            DataConnection._from_dict(_dict=data_connection) for data_connection in training_data_references]

        run_details = self.get_run_details(include_metrics=True, _internal=True)

        for data_connection in data_connections:  # note: populate DataConnections with optimizer params
            data_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            data_connection.set_client(self._engine._wml_client)
            data_connection._run_id = self._engine._current_run_id
            data_connection._user_holdout_exists = user_holdout_exists

            metrics = run_details['entity']['status'].get('metrics', [])
            if metrics and metrics[-1]['context'].get('fairness'):
                data_connection.auto_pipeline_params['fairness_info'] = metrics[-1]['context']['fairness'].get('info')

        return data_connections

    def get_test_data_connections(self) -> List['DataConnection']:
        """Create DataConnection objects for further user usage (To recreate autoai holdout that user specified).

        :return: list of DataConnection with populated optimizer parameters
        :rtype: list[DataConnection]
        """
        optimizer_parameters = self.get_params()
        run_details = self.get_run_details(_internal=True)

        if not run_details['entity'].get('test_data_references'):
            raise TestDataNotPresent(reason="User specified test data was not present in this experiment. "
                                            "Try to use 'with_holdout_split' parameter for original "
                                            "training_data_references to retrieve test data.")

        test_data_references = run_details['entity']['test_data_references']

        data_connections = [
            DataConnection._from_dict(_dict=data_connection) for data_connection in test_data_references]

        for data_connection in data_connections:  # note: populate DataConnections with optimizer params
            data_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            data_connection.set_client(self._engine._wml_client)
            data_connection._run_id = self._engine._current_run_id
            data_connection._test_data = True
            data_connection._user_holdout_exists = True

        return data_connections

    def _validate_training_data_size(self, training_data_reference: List['DataConnection']) -> None:
        """Check size of dataset in training data connection."""

        for training_connection in training_data_reference:
            t_shirt_size = self._engine._auto_pipelines_parameters.get('t_shirt_size')

            if isinstance(training_connection.location, S3Location):
                size = training_connection.location._get_file_size(training_connection._init_cos_client())
            elif isinstance(training_connection.location, WSDAssetLocation):
                size = training_connection.location._get_file_size()
            else:
                size = training_connection.location._get_file_size(self._workspace)

            if size is None:
                raise TrainingDataSourceIsNotFile(training_connection.location)

            if self._workspace.WMLS:
                if t_shirt_size in (TShirtSize.S, TShirtSize.M) and size > 100 * 1024 * 1024:
                    raise DataSourceSizeNotSupported()
                elif t_shirt_size in (TShirtSize.L) and size > 1024 * 1024 * 1024:
                    raise DataSourceSizeNotSupported()
            else:
                pass  # note: data source size checking for other environment is not implemented

    def get_preprocessed_data_connection(self) -> 'DataConnection':
        """Create DataConnection object for further user usage (with OBM output)
            (eg. to handle data storage connection or to recreate autoai holdout split).

        :return: DataConnection with populated optimizer parameters
        :rtype: DataConnection
        """
        optimizer_parameters = self.get_params()

        if optimizer_parameters['data_join_graph']:
            details = self._engine._wml_client.training.get_details(
                training_uid=self._engine._current_run_id, _internal=True)

            path = details['entity']['results_reference']['location'].get(
                'path',
                details['entity']['results_reference']['location'].get('file_name'))

            if path == '':
                path = f"{self._engine._current_run_id}/data/obm/features/part"

            else:
                path = f"{path}/{self._engine._current_run_id}/data/obm/features/part"

            results_connection = DataConnection._from_dict(_dict=details['entity']['results_reference'])
            results_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            results_connection.set_client(self._engine._wml_client)
            results_connection._run_id = self._engine._current_run_id
            results_connection._obm = True  # indicator for OBM output data
            results_connection._obm_cos_path = path

            return results_connection

        else:
            raise MissingDataPreprocessingStep(
                reason="Cannot get preprocessed data as preprocessing step was not performed.")

    def get_preprocessed_test_data_connection(self) -> 'DataConnection':
        """Create DataConnection object for further user usage (with OBM output)
            (eg. to handle data storage connection or to recreate autoai holdout split).

        :return: DataConnection with populated optimizer parameters
        :rtype: DataConnection
        """
        optimizer_parameters = self.get_params()

        if optimizer_parameters['data_join_graph']:
            details = self._engine._wml_client.training.get_details(
                training_uid=self._engine._current_run_id, _internal=True)

            path = details['entity']['test_output_data']['location'].get(
                'path',
                details['entity']['test_output_data']['location'].get('file_name'))

            if path == '':
                path = f"{self._engine._current_run_id}/data/obm/features/part"

            else:
                path = f"{path}/{self._engine._current_run_id}/data/obm/features/part"

            results_connection = DataConnection._from_dict(_dict=details['entity']['test_output_data'])
            results_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            results_connection.set_client(self._engine._wml_client)
            results_connection._run_id = self._engine._current_run_id
            results_connection._obm = True  # indicator for OBM output data
            results_connection._obm_cos_path = path

            return results_connection

        else:
            raise MissingDataPreprocessingStep(
                reason="Cannot get preprocessed data as preprocessing step was not performed.")

    def get_preprocessing_pipeline(self) -> 'DataJoinPipeline':
        """Returns preprocessing pipeline object for further usage. (eg. to visualize preprocessing pipeline as graph).

        :return: preprocessing pipeline object
        :rtype: DataJoinPipeline
        """
        optimizer_parameters = self.get_params()

        if optimizer_parameters['data_join_graph']:
            return DataJoinPipeline(preprocessed_data_connection=self.get_preprocessed_data_connection(),
                                    optimizer=self)

        else:
            raise MissingDataPreprocessingStep(
                reason="Cannot get preprocessed pipeline as preprocessing step was not performed.")



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import ibm_watson_machine_learning._wrappers.requests as requests
import json
import base64
import logging
from datetime import datetime, timedelta
from ibm_watson_machine_learning.wml_client_error import NoWMLCredentialsProvided, ApiRequestFailure, WMLClientError, \
                                                         CannotAutogenerateBedrockUrl
from ibm_watson_machine_learning.href_definitions import HrefDefinitions


class ServiceInstanceNewPlan:
    """Connect, get details and check usage of Watson Machine Learning service instance."""

    def __init__(self, client):
        self._logger = logging.getLogger(__name__)
        self._client = client
        self._wml_credentials = client.wml_credentials
        self._expiration_datetime = None

        if self._client.ICP_PLATFORM_SPACES:
            if self.get_instance_id() == 'openshift' or self.get_instance_id() == 'wml_local':
                self._wml_credentials[u'url'] = self.get_url()
            else:
                self._wml_credentials[u'url'] = self.get_url() + ':31843'
            # TODO: Check if this is used anywhere.. from initial searches, doesn't seem like
            # self._wml_credentials[u'instance_id'] = "999"

        # This is used in connections.py
        self._href_definitions = HrefDefinitions(self._client,
                                                 self._client.CLOUD_PLATFORM_SPACES,
                                                 self._client.PLATFORM_URL,
                                                 self._client.CAMS_URL,
                                                 self._client.ICP_PLATFORM_SPACES)

        self._client.wml_token = self._create_token()
        # self._logger.info(u'Successfully prepared token: ' + self._client.wml_token)
        # ml_repository_client is initialized in repo
        self.details = None

    def get_instance_id(self):
        """Get instance id of Watson Machine Learning service.

        :return: instance id
        :rtype: str

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_instance_id()
        """
        if self._wml_credentials['instance_id'] == 'invalid':
            raise WMLClientError('instance_id for this plan is picked up from the space with which'
                                 'this instance_id is associated with. Set the space with associated'
                                 'instance_id to be able to use this function')
        return self._wml_credentials['instance_id']

    def get_api_key(self):
        """Get api key of Watson Machine Learning service.

        :return: api key
        :rtype: str

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_api_key()
        """
        return self._wml_credentials['apikey']

    def get_url(self):
        """Get instance url of Watson Machine Learning service.

        :return: instance url
        :rtype: str

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_url()
        """
        return self._wml_credentials['url']

    def get_username(self):
        """Get username for Watson Machine Learning service. Applicable only for IBM Cloud Pak® for Data.

        :return: username
        :rtype: str

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_username()
        """
        if self._client.ICP_PLATFORM_SPACES:
            try:
                return self._wml_credentials['username']
            except:
                raise WMLClientError('`username` missing in wml_credentials.')
        else:
            raise WMLClientError("Not applicable for Cloud")

    def get_password(self):
        """Get password for Watson Machine Learning service. Applicable only for IBM Cloud Pak® for Data.

        :return: password
        :rtype: str

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_password()
        """
        if self._client.ICP_PLATFORM_SPACES:
            try:
                return self._wml_credentials['password']
            except:
                raise WMLClientError('`password` missing in wml_credentials.')
        else:
            raise WMLClientError("Not applicable for Cloud")

    def get_details(self):
        """Get information about Watson Machine Learning instance.

        :return: metadata of service instance
        :rtype: dict

        **Example**

        .. code-block:: python

            instance_details = client.service_instance.get_details()
        """

        if not self._client.ICP:
            if self._wml_credentials is not None:

                if self._wml_credentials['instance_id'] == 'invalid':
                    raise WMLClientError('instance_id for this plan is picked up from the space with which '
                                         'this instance_id is associated with. Set the space with associated '
                                         'instance_id to be able to use this function')

                    # /ml/v4/instances will need either space_id or project_id as mandatory params
                # We will enable this service instance class only during create space or
                # set space/project. So, space_id/project_id would have been populated at this point
                headers = self._client._get_headers()

                del headers[u'X-WML-User-Client']
                if 'ML-Instance-ID' in headers:
                    headers.pop('ML-Instance-ID')
                headers.pop(u'x-wml-internal-switch-to-new-v4')
                # params = {'version': self._client.version_param}
                response_get_instance = requests.get(
                    self._href_definitions.get_v4_instance_id_href(),
                    params=self._client._params(skip_space_project_chk=True),
                    # params={'version': self._client.version_param},
                    headers=headers
                )

                if response_get_instance.status_code == 200:
                    return response_get_instance.json()
                else:
                    raise ApiRequestFailure(u'Getting instance details failed.', response_get_instance)
            else:
                raise NoWMLCredentialsProvided
        else:
            return {}

    def _get_token(self):
        if self._client.wml_token is None:
            self._create_token()
            self._client.repository._refresh_repo_client()
        elif self._is_token_refresh_possible():
            if self._client.ICP:
                if self._get_expiration_datetime():
                    if self._get_expiration_datetime() - timedelta(minutes=50) < datetime.now():
                        self._client.wml_token = self._get_cpd_token_from_request()
                        self._client.repository._refresh_repo_client()
                else:
                    self._client.wml_token = self._get_cpd_token_from_request()
                    self._client.repository._refresh_repo_client()
            elif self._client._is_IAM():
                if self._get_expiration_datetime() - timedelta(minutes=15) < datetime.now():
                    self._client.wml_token = self._get_IAM_token()
                    self._client.repository._refresh_repo_client()
            elif self._get_expiration_datetime() - timedelta(minutes=30) < datetime.now():
                self._client.repository._refresh_repo_client()
                self._refresh_token()

        return self._client.wml_token

    def _create_token(self):

        if self._client.proceed is True:
            return self._wml_credentials["token"]

        if not self._client.ICP_PLATFORM_SPACES:
            if self._client._is_IAM():
                return self._get_IAM_token()
            else:
                raise WMLClientError('apikey for IAM token is not provided in credentials for the client.')
        else:
            return self._get_cpd_token_from_request()

    def _refresh_token(self):
        if self._client.proceed is True:
            self._client.wml_token = self._wml_credentials["token"]

        self._client.wml_token = self._get_cpd_token_from_request()

    def _get_expiration_datetime(self):
        if self._expiration_datetime is not None:
            return self._expiration_datetime

        token_parts = self._client.wml_token.split('.')
        token_padded = token_parts[1] + '=' * (len(token_parts[1]) % 4)
        token_info = json.loads(base64.b64decode(token_padded).decode('utf-8'))
        token_expire = token_info.get('exp')

        return datetime.fromtimestamp(token_expire)

    def _is_iam(self):
        try:
            token_parts = self._client.wml_token.split('.')
            token_padded = token_parts[1] + '=' * (len(token_parts[1]) % 4)
            token_info = json.loads(base64.b64decode(token_padded).decode('utf-8'))
            instanceId = token_info.get('instanceId')

            return instanceId
        except:
            return False

    def _get_IAM_token(self):
        if self._client.proceed is True:
            return self._wml_credentials["token"]
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Basic Yng6Yng='
        }

        mystr = 'apikey=' + self._href_definitions.get_iam_token_api()
        response = requests.post(
            self._href_definitions.get_iam_token_url(),
            data=mystr,
            headers=headers
        )

        if response.status_code == 200:
            token = response.json().get(u'access_token')
            self._expiration_datetime = None
        else:
            raise WMLClientError(u'Error getting IAM Token.', response)
        return token

    def _is_token_refresh_possible(self):
        """Checks if necessary credentials were passed for token refresh.
        For CP4D we need (username & password)/(username & api_key).
        For Cloud we need api_key.

        :return: `True` if token refresh can be performed `False` otherwise
        :rtype: bool
        """
        if self._client._is_IAM():
            return 'apikey' in self._wml_credentials
        else:
            return 'username' in self._wml_credentials and (
                    'password' in self._wml_credentials or 'apikey' in self._wml_credentials)

    def _get_cpd_auth_pair(self):
        """Get a pair of credentials required for generation of token.

        :return: string representing a dictionary of authentication credentials
                         (username & password) or (username & api_key).
        :rtype: str
        """
        if "apikey" in self._wml_credentials:
            return f'{{\"username\": \"{self.get_username()}\", \"api_key\": \"{self.get_api_key()}\"}}'
        else:
            return f'{{\"username\": \"{self.get_username()}\", \"password\": \"{self.get_password()}\"}}'

    def _get_cpd_bedrock_auth_data(self):
        """Get data required for generation of token.

        :return: string representing a dictionary of authentication credentials
        :rtype: str
        """
        return f'grant_type=password&username={self.get_username()}&password={self.get_password()}&scope=openid'

    def _get_cpd_token_from_request_old_auth_flow(self):
        token_url = self._href_definitions.get_cpd_token_endpoint_href()
        response = requests.post(token_url,
                                 headers={
                                     'Content-Type': 'application/json'
                                 },
                                 data=self._get_cpd_auth_pair())

        if response.status_code == 200:
            self._expiration_datetime = None
            return response.json().get(u'token')
        else:
            raise WMLClientError(u'Error refreshing the token.', response)

    def _get_cpd_token_from_request_new_auth_flow(self):
        bedrock_url = self._href_definitions.get_cpd_bedrock_token_endpoint_href()
        response = requests.post(bedrock_url,
                                 headers={
                                     'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8'
                                 },
                                 data=self._get_cpd_bedrock_auth_data())

        if response.status_code != 200:
            raise WMLClientError(u'Error refreshing the token.', response, logg_messages=False)

        iam_token = response.json()['access_token']
        self._expiration_datetime = datetime.now() + timedelta(seconds=response.json()['expires_in'])
        # refresh_token = response.json()['refresh_token']

        token_url = self._href_definitions.get_cpd_validation_token_endpoint_href()
        response = requests.get(token_url,
                                headers={
                                    'username': self.get_username(),
                                    'iam-token': iam_token
                                })

        if response.status_code == 200:
            return response.json()['accessToken']
        else:
            raise WMLClientError(u'Error refreshing the token.', response)

    def _get_cpd_token_from_request(self):
        """Send a request for token on CPD.

        :return: newly created token is returned if no errors occurred
        :rtype: str
        """
        if (self._client.ICP_40 or self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47) and 'bedrock_url' in self._client.wml_credentials and 'password' in self._client.wml_credentials:
            try:
                return self._get_cpd_token_from_request_new_auth_flow()
            except Exception as e1:
                if not hasattr(self._client, '_is_bedrock_url_autogenerated'):
                    raise e1

                try:
                    res = self._get_cpd_token_from_request_old_auth_flow()
                    # if it worked then iamintegration=False, then removing bedrock_url will shorten the path
                    del self._client.wml_credentials['bedrock_url']
                    return res
                except Exception as e2:
                    if hasattr(self._client, '_is_bedrock_url_autogenerated') and self._client._is_bedrock_url_autogenerated:
                        raise CannotAutogenerateBedrockUrl(e1, e2)
                    else:
                        raise e2
        else:
            return self._get_cpd_token_from_request_old_auth_flow()



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class MlAssetsCreateExperimentInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, tags=None, name=None, description=None, framework=None, training_data_reference=None, author=None):
        """
        MlAssetsCreateExperimentInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'tags': 'list[TagRepository]',
            'name': 'str',
            'description': 'str',
            'framework': 'FrameworkOutputRepository',
            'training_data_reference': 'ArrayDataInputRepository',
            'author': 'AuthorRepository'
        }

        self.attribute_map = {
            'tags': 'tags',
            'name': 'name',
            'description': 'description',
            'framework': 'framework',
            'training_data_reference': 'training_data_reference',
            'author': 'author'
        }
        self._tags = tags
        self._name = name
        self._description = description
        self._framework = framework
        self._training_data_reference = training_data_reference
        self._author = author

    @property
    def tags(self):
        """
        Gets the tags of this ExperimentInputExperiments.


        :return: The tags of this ExperimentInputExperiments.
        :rtype: list[TagRepository]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of this ExperimentInputExperiments.


        :param tags: The tags of this ExperimentInputExperiments.
        :type: list[TagRepository]
        """

        self._tags = tags

    @property
    def name(self):
        """
        Gets the name of this MlAssetsCreateExperimentInput.


        :return: The name of this MlAssetsCreateExperimentInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this MlAssetsCreateExperimentInput.


        :param name: The name of this MlAssetsCreateExperimentInput.
        :type: str
        """
        self._name = name

    @property
    def description(self):
        """
        Gets the description of this MlAssetsCreateExperimentInput.


        :return: The description of this MlAssetsCreateExperimentInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this MlAssetsCreateExperimentInput.


        :param description: The description of this MlAssetsCreateExperimentInput.
        :type: str
        """
        self._description = description

    @property
    def framework(self):
        """
        Gets the framework of this MlAssetsCreateExperimentInput.


        :return: The framework of this MlAssetsCreateExperimentInput.
        :rtype: FrameworkOutputRepository
        """
        return self._framework

    @framework.setter
    def framework(self, framework):
        """
        Sets the framework of this MlAssetsCreateExperimentInput.


        :param framework: The framework of this MlAssetsCreateExperimentInput.
        :type: FrameworkOutputRepository
        """
        self._framework = framework

    @property
    def training_data_reference(self):
        """
        Gets the training_data of this MlAssetsCreateExperimentInput.


        :return: The training_data of this MlAssetsCreateExperimentInput.
        :rtype: ArrayDataInputRepository
        """
        return self._training_data_reference

    @training_data_reference.setter
    def training_data_reference(self, training_data_reference):
        """
        Sets the training_data of this MlAssetsCreateExperimentInput.


        :param training_data: The training_data of this MlAssetsCreateExperimentInput.
        :type: ArrayDataInputRepository
        """
        self._training_data_reference = training_data_reference

    @property
    def author(self):
        """
        Gets the author of this MlAssetsCreateExperimentInput.


        :return: The author of this MlAssetsCreateExperimentInput.
        :rtype: AuthorRepository
        """
        return self._author

    @author.setter
    def author(self, author):
        """
        Sets the author of this MlAssetsCreateExperimentInput.


        :param author: The author of this MlAssetsCreateExperimentInput.
        :type: AuthorRepository
        """
        self._author = author

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#!/usr/bin/env python3

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import requests as req
import os
import sys
import ssl
import importlib.util
from pathlib import Path
import platform
import ibm_watson_machine_learning._wrappers.requests as requests
import json
import logging
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.utils.utils import get_module_version

logger = logging.getLogger(__name__)


def process_heartbeat(client, **kwargs):
    try:
        config_dict = kwargs.get('config_dict')
        agg_info = config_dict.get('aggregator')
        wml_services_url = agg_info['ip'].split('/')[0]

        hearbeat_resp = requests.get("https://" + wml_services_url + "/wml_services/training/heartbeat")
        hearbeat_resp_json = json.loads(hearbeat_resp.content.decode("utf-8"))
        service_version = hearbeat_resp_json.get("service_implementation")

        if service_version.startswith("4.7"):
            return "cpd470"
        if service_version.startswith("4.6"):
            return "cpd460"
        if service_version.startswith("4.5") or service_version.startswith("45") or service_version.startswith("0.1.10000"):
            return "cpd450"
        if service_version.startswith("4.0"):
            version_details = service_version.split('-')
            if version_details[1] >= "202203":
                return "cpd408"
            if version_details[1] >= "202202":
                return "cpd407"
            if version_details[1] >= "202112":
                return "cpd406"
            elif version_details[1] >= "202110":
                return "cpd403"
            elif version_details[1] >= "202108":
                return "cpd402"
            else:
                return "cpd401"
        if service_version.startswith("35"):
            return "cpd35"

        return "cloud"

    except Exception as ex:
        logger.info("unable to process heartbeat")
        logger.exception(ex)
        return "cloud"


def import_diff(module_file_path):

    if "py38-tf24-pt17-sk23" in module_file_path:
        return
    pathlist = Path(module_file_path).rglob('*.py')
    for path in pathlist:
        # because path is object not string
        path_in_str = str(path)
        if not path_in_str.endswith("__init__.py"):
            module_name = ("ibmfl" + path_in_str.split("ibmfl")[2].replace(os.sep, "."))[:-3]
            module_spec = importlib.util.spec_from_file_location(
                module_name, path_in_str)
            loader = importlib.util.LazyLoader(module_spec.loader)
            module_spec.loader = loader
            module = importlib.util.module_from_spec(module_spec)
            sys.modules[module_name] = module
            module_spec.loader.exec_module(module)


def check_python_framework_version(client_reqs):
    client_system = platform.system()
    client_processor = platform.processor()
    
    if 'fl_extras' in client_reqs:
        install_msg = "  You can install this and other required packages by running pip install --upgrade 'ibm-watson-machine-learning[" + client_reqs['fl_extras'] + "]'"
    else:
        install_msg = "  See documentation for more information."
        
    py_version = platform.python_version()
    logger.info("Detected Client Python Version: {}".format(py_version))
    if not py_version.startswith(client_reqs['py_version']):
            raise Exception("The selected software spec requires python=={}.".format(client_reqs['py_version']))

    if client_system == "Darwin" and client_processor == "arm":
        try:
            tensorflowmacos_version = get_module_version(lib_name='tensorflow-macos')
            logger.info("Detected tensorflow-macos Version: {}".format(tensorflowmacos_version))
            if not tensorflowmacos_version.startswith(client_reqs['tensorflow_version']):
                raise Exception("Incompatible tensorflow-macos version found")
        except Exception as ex:
            logger.warning("{}, this may cause unexpected errors.".format(ex))
            logger.warning("The selected software spec requires tensorflow-macos=={}.".format(client_reqs['tensorflow_version']))
    else:
        try:
            tensorflow_version = get_module_version(lib_name='tensorflow')
            logger.info("Detected tensorflow Version: {}".format(tensorflow_version))
            if not tensorflow_version.startswith(client_reqs['tensorflow_version']):
                raise Exception("Incompatible tensorflow version found")
        except Exception as ex:
            logger.warning("{}, this may cause unexpected errors.".format(ex))
            logger.warning("The selected software spec requires tensorflow=={}.".format(client_reqs['tensorflow_version']) + install_msg)

    try:
        torch_version = get_module_version(lib_name='torch')
        logger.info("Detected torch Version: {}".format(torch_version))
        if not torch_version.startswith(client_reqs['torch_version']):
            raise Exception("Incompatible torch version found")
    except Exception as ex:
        logger.warning("{}, this may cause unexpected errors.".format(ex))
        logger.warning("The selected software spec requires torch=={}.".format(client_reqs['torch_version']) + install_msg)
    
    try:
        scikitlearn_version = get_module_version(lib_name='scikit-learn')
        logger.info("Detected scikit-learn Version: {}".format(scikitlearn_version))
        if not scikitlearn_version.startswith(client_reqs['scikitlearn_version']):
            raise Exception("Incompatible scikit-learn version found")
    except Exception as ex:
        logger.warning("{}, this may cause unexpected errors.".format(ex))
        logger.warning("The selected software spec requires scikit-learn=={}.".format(client_reqs['scikitlearn_version']) + install_msg)


def choose_software_version(software_spec):
    logger.info("Aggregator Software Spec: {}".format(software_spec))
    
    if software_spec == "runtime-22.1-py3.9" or software_spec == "12b83a17-24d8-5082-900f-0ab31fbfd3cb":
        client_reqs = {'py_version': '3.9', 'tensorflow_version': '2.7', 'torch_version': '1.10', 'scikitlearn_version': '1.0', 'fl_extras': 'fl-rt22.1'}
        check_python_framework_version(client_reqs)
        return "py39-tf27-pt110-sk10"
    elif software_spec == "runtime-22.2-py3.10" or software_spec == "b56101f1-309d-549b-a849-eaa63f77b2fb":
        client_reqs = {'py_version': '3.10', 'tensorflow_version': '2.9', 'torch_version': '1.12', 'scikitlearn_version': '1.1', 'fl_extras': 'fl-rt22.2-py3.10'}
        check_python_framework_version(client_reqs)
        return "py310-tf29-pt112-sk11"
    elif software_spec == "runtime-23.1-py3.10" or software_spec == "336b29df-e0e1-5e7d-b6a5-f6ab722625b2":
        client_reqs = {'py_version': '3.10', 'tensorflow_version': '2.12', 'torch_version': '2.0', 'scikitlearn_version': '1.1', 'fl_extras': 'fl-rt23.1-py3.10'}
        check_python_framework_version(client_reqs)
        return "py310-tf212-pt20-sk11"
    else:
        client_reqs = {'py_version': '3.8', 'tensorflow_version': '2.4', 'torch_version': '1.7', 'scikitlearn_version': '0.23', 'fl_extras': 'fl'}
        check_python_framework_version(client_reqs)
        return "py38-tf24-pt17-sk23"


fl_path = os.path.abspath('.')
if fl_path not in sys.path:
    sys.path.append(fl_path)


class Party(WMLResource):
    """The Party class embodies a Federated Learning party, with methods to run, cancel, and query local training.
    Refer to the ``client.remote_training_system.create_party()`` API for more information about creating an
    instance of the Party class. 
    """

    SUPPORTED_PLATFORMS_MAP = {
        'cloud': "/cloud/ibmfl",
        'cpd35': "1.0.181",
        'cpd401': "1.0.122",
        'cpd402': "1.0.165",
        'cpd403': "/cpd403/ibmfl",
        'cpd406': "/cpd406/ibmfl",
        'cpd407': "/cpd406/ibmfl",
        'cpd408': "/cpd406/ibmfl",
        'cpd450': "/cpd406/ibmfl",
        'cpd460': "/cpd406/ibmfl",
        'cpd470': "/cpd406/ibmfl",
        'py38-tf24-pt17-sk23': "/py38-tf24-pt17-sk23/ibmfl", #default
        'py39-tf27-pt110-sk10': "/py39-tf27-pt110-sk10/ibmfl",
        'py310-tf29-pt112-sk11': "/py310-tf29-pt112-sk11/ibmfl",
        'py310-tf212-pt20-sk11': "/py310-tf212-pt20-sk11/ibmfl",
    }

    def __init__(self, client=None, **kwargs):

        libs_module = sys.modules['ibm_watson_machine_learning.libs']
        libs_location_list = libs_module.__path__

        # base location string, default to cloud location
        ibmfl_base_module_location = libs_location_list[0] + '/ibmfl'

        # get arg for platform
        calculated_platform_env = process_heartbeat(client=client, **kwargs)
        platform_env = kwargs.get('env', calculated_platform_env)
        logger.info("Detected platform environment: {}".format(platform_env))

        # process location
        ibmfl_module_location = ibmfl_base_module_location + self.SUPPORTED_PLATFORMS_MAP.get(platform_env)
        if not ibmfl_module_location.endswith("ibmfl"):
            raise Exception("Please use ibm-watson-machine-learning " + self.SUPPORTED_PLATFORMS_MAP.get(platform_env) )

        #check if using old connector script which is removed
        if not client:
            raise Exception("This version of the party connector script is outdated. "
                            "Please download the party connector script from your current Federated Learning experiment. "
                            "For more details, please refer to the documentation.")

        self.module_location = ibmfl_module_location
        self.args = kwargs
        self.Party = None
        self.connection = None
        self.log_level = None
        self.metrics_output = None

        if 'ibmfl' in sys.modules:
            del sys.modules['ibmfl']
        if 'ibmfl.party' in sys.modules:
            del sys.modules['ibmfl.party']
        if 'ibmfl.party.party' in sys.modules:
            del sys.modules['ibmfl.party.party']

        #install the general lib (which is the cloud version)
        module_name = 'ibmfl'
        module_spec = importlib.util.spec_from_file_location(
            module_name, ibmfl_base_module_location + '/py38-tf24-pt17-sk23/ibmfl/__init__.py')
        module = importlib.util.module_from_spec(module_spec)
        sys.modules[module_name] = module
        module_spec.loader.exec_module(module)

        WMLResource.__init__(self, __name__, client)
        self._client = client
        self.auth_token = "Bearer " + self._client.wml_token
        self.project_id = self._client.project_id
        self.log_level = kwargs.get("log_level", "ERROR")

    def start(self):
        self.Party.start()

    def run(self, aggregator_id=None, experiment_id=None, asynchronous=True, verify=True, timeout: int = 60 * 10):
        """Connect to a Federated Learning aggregator and run local training.
        Exactly one of `aggregator_id` and `experiment_id` must be supplied.

        :param aggregator_id: aggregator identifier

            * If aggregator_id is supplied, the party will connect to the given aggregator.

        :type aggregator_id: str, optional
        :param experiment_id: experiment identifier

            * If experiment_id is supplied, the party will connect to the most recently created aggregator
                for the experiment.

        :type experiment_id: str, optional
        :param asynchronous:
            * `True`  - party starts to run the job in the background and progress can be checked later
            * `False` - method will wait until training is complete and then print the job status
        :type asynchronous: bool, optional
        :param verify: verify certificate
        :type verify: bool, optional
        :param timeout: timeout in seconds

              * If the aggregator is not ready within timeout seconds, exit.

        :type timeout: int, or None for no timeout

        **Examples**

        .. code-block:: python

            party.run( aggregator_id = "69500105-9fd2-4326-ad27-7231aeb37ac8", asynchronous = True, verify = True )
            party.run( experiment_id = "2466fa06-1110-4169-a166-01959adec995", asynchronous = False )

        """
        from ibmfl.exceptions import FLException
        from datetime import datetime
        import time

        timeout_time = None if timeout is None else timeout + time.time()
        if ( experiment_id is None and aggregator_id is None ) or ( experiment_id is not None and aggregator_id is not None ) :
            raise FLException("Exactly one of aggregator_id and experiment_id must be supplied")

        if ( experiment_id is not None ):
            while True: 
                try:
                    details = self._client.training.get_details(get_all=True,training_definition_id=experiment_id, _internal=True)['resources']
                    details = [ d for d in details if d['entity']['status']['state'] in ['accepting_parties','pending','running'] ]
                    if not details :
                        if timeout_time and timeout_time < time.time() :
                            raise FLException("Cannot find an aggregator for experiment %s" % experiment_id)
                        else :
                            logger.info("Cannot find an aggregator for experiment %s . Retrying." % experiment_id)
                            time.sleep(30)
                            continue
                    else :
                        aggregator_id = max( [(t['metadata']['id'],datetime.strptime(t['metadata']['created_at'],'%Y-%m-%dT%H:%M:%S.%fZ'))
                                          for t in details],key=lambda d:d[1])[0]
                        logger.info("Using aggregator id %s" % aggregator_id) 
                        break
                except Exception as ex:
                    logger.exception(ex)
                    raise FLException(str(ex))


        #import the changed files in the desired lib version

        config_dict = self.args.get('config_dict')
        metrics_config = {
            "name": "WMLMetricsRecorder",
            "path": "ibmfl.party.metrics.metrics_recorder",
            "output_file": self.metrics_output,
            "output_type": "json",
            'compute_pre_train_eval': False,
            'compute_post_train_eval': False
        }
        if 'metrics_recorder' not in config_dict:
            config_dict["metrics_recorder"] = metrics_config
        wml_services_url = config_dict.get('aggregator').get('ip').split('/')[0]
        agg_info = wml_services_url + "/ml/v4/trainings/" + aggregator_id
        config_dict['aggregator']['ip'] = agg_info
        self.args['config_dict'] = config_dict

        #verify ssl context
        if verify:
            try:
                req.get("https://" + wml_services_url + "/wml_services/training/heartbeat", verify=verify)
            except requests.exceptions.SSLError as ex:
                logger.error(str(ex))
                raise FLException("No valid certificate detected. Please replace the default certificate with your own "
                                  "TLS certificate, or set verify to False at your own risk. For more details, please see "
                                  "https://www.ibm.com/docs/en/cloud-paks/cp-data/4.0?topic=client-using-custom-tls-certificate-connect-platform")

        #check for aggregator state and start job
        try:
            training_status = self._client.training.get_status(aggregator_id)
            state = training_status['state']
            ready = False
            if state == "pending":
                while state == "pending" and ( not timeout_time or timeout_time > time.time() ) :
                    logger.info("Waiting for aggregator accepting parties state..")
                    time.sleep(10)
                    training_status = self._client.training.get_status(aggregator_id)
                    state = training_status['state']
                if state != "accepting_parties":
                    raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))
                ready = True
            elif state == "running" or state == "accepting_parties":
                ready = True
            else:
                raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))

            if ready:
                if "cpd406" in self.module_location or "cloud" in self.module_location:
                    details = self._client.training.get_details(aggregator_id, _internal=True)
                    fl_entity = details["entity"]["federated_learning"]
                    if "software_spec" in fl_entity:
                        software_spec = fl_entity["software_spec"]["name"] if "name" in fl_entity["software_spec"] else fl_entity["software_spec"]["id"]
                    else:
                        software_spec = "runtime-22.1-py3.9"
                    platform_env = choose_software_version(software_spec)
                    logger.info("Loading {} environment..".format(platform_env))
                    if "system" in details and "warnings" in details["system"]:
                        for warning in details["system"]["warnings"]:
                            logger.info("Warning: {}".format(warning["message"]))
                    self.module_location = '/'.join(self.module_location.split('/')[0:-2]) + self.SUPPORTED_PLATFORMS_MAP.get(platform_env)
                import_diff(self.module_location)
                from ibmfl.party.party import Party
                self.Party = Party(**self.args, token=self.auth_token, self_signed_cert=not verify, log_level=self.log_level)
                self.connection = self.Party.connection
                self.start()
            else:
                raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))
            #wait for the job to finish if synchrounous
            if not asynchronous:
                while "completed" != state and "failed" != state and "canceled" != state and self.is_running():
                    training_status = self._client.training.get_status(aggregator_id)
                    state = training_status['state']
                    time.sleep(10)
                logger.info("The training finishes with %s status" % state)
        except FLException as ex:
            raise FLException(str(ex))
        except Exception as ex:
            logger.info("The party failed to start training")
            logger.exception(ex)

    def monitor_logs(self, log_level="INFO"):
        """Enable logging of the training job to standard output.
        This method should be called before calling the ``run()`` method.

        :param log_level: log level specified by user
        :type log_level: str, optional

        **Example**

        .. code-block:: python

            party.monitor_logs()

        """
        self.log_level = log_level

        # Configure logging locally as well
        from ibmfl.util.config import configure_logging_from_file
        configure_logging_from_file(log_level=log_level)
        logger.setLevel(log_level)

    def monitor_metrics(self, metrics_file="-"):
        """Enable output of training metrics.

        :param metrics_file: a filename specified by user to which the metrics should be written
        :type metrics_file: str, optional

        .. note::
            This method outputs the metrics to stdout if a filename is not specified

        **Example**

        .. code-block:: python

            party.monitor_metrics()

        """
        self.metrics_output = metrics_file

    def is_running(self):
        """Check if the training job is running.

        :return: if the job is running
        :rtype: bool

        **Example**

        .. code-block:: python

            party.is_running()

        """
        return not self.connection.stopped

    def get_round(self):
        """Get the current round number.

        :return: the current round number
        :rtype: int

        **Example**

        .. code-block:: python

            party.get_round()

        """
        return self.Party.proto_handler.metrics_recorder.get_round_no()

    def cancel(self):
        """Stop the local connection to the training on the party side.

        **Example**

        .. code-block:: python

            party.cancel()

        """
        self.Party.stop_connection()



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import torch
import logging
import time
import copy
from importlib import import_module

import numpy as np
from sklearn import metrics
from skorch import NeuralNet
from skorch.callbacks import EpochScoring
from skorch.exceptions import NotInitializedError

from ibmfl.model.fl_model import FLModel
from ibmfl.model.model_update import ModelUpdate
from ibmfl.util import config
from ibmfl.exceptions import LocalTrainingException, ModelException, FLException

logger = logging.getLogger(__name__)


class PytorchFLModel(FLModel):
    """
    Wrapper class for importing a Pytorch based model
    """

    def __init__(self, model_name,
                 model_spec=None,
                 pytorch_module=None,
                 module_init_params=None,
                 **kwargs):
        """
        Create a `PytorchFLModel` instance from a Pytorch model.
        If pytorch_model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: String specifying the type of model e.g., Pytorch_NN
        :type model_name: `str`
        :param model_spec: A dictionary specifying path to saved nn.sequence container
        :type model_spec: 'dict'
        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference \
        :param module_init_params: A dictionary with the values for the model's init \
        arguments. The key for the init parameters must be prefixed with module__ \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a PyTorch model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec, **kwargs)

        self.model_type = 'PyTorch'
        self.model = None
        self.module = None
        self.optimizer = torch.optim.SGD
        self.criterion = torch.nn.NLLLoss
        if pytorch_module is None:
            if model_spec is None or (not isinstance(model_spec, dict)):
                raise ValueError('Initializing model requires '
                                 'a model specification or uninstantiated '
                                 'pytorch model class reference '
                                 'None was provided')
            # In this case we need to recreate the model from model_spec
            self.module = self.load_model_from_spec(model_spec)
        else:
            self.module = pytorch_module
        self.model = self.initialize_model(self.module, self.optimizer,
                                           self.criterion, module_init_params)

        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            if self.num_gpus > torch.cuda.device_count():
                logger.error('Selected number of gpus to use for training exceeds number of available gpus, ' +
                             str(torch.cuda.device_count()) +
                             'Set number of gpus to maximum available on device.')
                self.num_gpus = torch.cuda.device_count()

            device_ids = list(range(self.num_gpus))
            self.model.module_ = torch.nn.DataParallel(self.model.module_, device_ids=device_ids)
            self.model.set_params(device='cuda')
            self.model.module_.to('cuda')

    def initialize_model(self, pytorch_module, optimizer, criterion,
                         module_init_params=None):
        """
        Initializes a pytorch model via skorch library

        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference
        :param optimizer: the optimizer to use
        :type optimizer: pytorch optimizer class
        :param criterion: the loss function to use
        :type criterion: pytorch loss function class
        :param module_init_params: A dictionary with the values for the model's init arguments. \
        The key for the init parameters must be prefixed with module__  \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :return: an initialized skorch model
        :rtype: 'skorch.NeuralNet'
        """
        if module_init_params is None:
            module_init_params = {}
        model = NeuralNet(
            module=pytorch_module,
            optimizer=optimizer,
            criterion=criterion,
            warm_start=True,
            callbacks=[('valid_acc', EpochScoring(self.valid_acc,
                                                  lower_is_better=False,
                                                  use_caching=True,
                                                  name='test_valid_acc'))],
            **module_init_params,
        )
        model.initialize()
        return model

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data, a tuple \
        given in the form (x_train, y_train). otherwise, input compatible with skorch.dataset.Dataset
        :type train_data: `(np.ndarray, ”np.ndarray)`
        :param fit_params: (optional) Dictionary with hyperparameters \
        that will be used to call fit function. \
        Hyperparameter parameters should match pytorch expected values \
        e.g., `epochs`, which specifies the number of epochs to be run. \
        If no `epochs` or `batch_size` are provided, a default value \
        will be used (1 and 128, respectively).
        :type fit_params: `dict`
        :return: None
        """
        # Initialized with default values
        batch_size = 128
        epochs = 1
        lr = 0.01
        optimizer_params = {}
        hyperparams = fit_params.get('hyperparams', {}) if fit_params else {}

        if hyperparams:
            local_params = hyperparams.get('local', {}) or {}
            training_hp = local_params.get('training', {}) or {}
            
            epochs = training_hp.get('epochs', epochs)
            batch_size = training_hp.get('batch_size', batch_size)
            lr = training_hp.get('lr', lr)
            
            if 'optimizer' in local_params:
                optimizer = local_params['optimizer']
                if isinstance(optimizer, str):
                    try:
                        optimizer = getattr(import_module(
                            'torch.optim'), optimizer.split('.')[-1])

                    except Exception as e:
                        optimizer = self.optimizer
                        logger.exception(str(e))
                        print('selected optimizer not found. using default')
                if optimizer != self.optimizer:
                    self.optimizer = optimizer
                    self.model.set_params(optimizer=self.optimizer)

                    if self.use_gpu_for_training and torch.cuda.device_count() > 0:
                        self.model.module_.to(self.model.device)

            if 'criterion' in local_params:
                criterion = local_params['criterion']
                if isinstance(criterion, str):
                    try:
                        criterion = getattr(import_module(
                            'torch.nn'), criterion.split('.')[-1])

                    except Exception as e:
                        criterion = self.criterion
                        logger.exception(str(e))
                        print('selected criterion not found. using default')

                if criterion != self.criterion:
                    self.criterion = criterion
                    self.model.set_params(criterion=self.criterion)

            if 'optimizer_params' in local_params:
                optimizer_params = local_params['optimizer_params']
        self.model.set_params(batch_size=batch_size, lr=lr, **optimizer_params)

        try:
            if type(train_data) is tuple:
                # Extract x_train and y_train, by default,
                # label is stored in the last column
                x = train_data[0]
                y = train_data[1]
                self.model.fit_loop(x, y, epochs=epochs)

            else:
                # otherwise, expect that input is a pytorch Dataset generator
                self.model.fit_loop(train_data, epochs=epochs)

        except Exception as e:
            logger.exception(str(e))
            raise LocalTrainingException(
                'Error occurred while performing model.fit')

    def update_model(self, model_update):
        """
        Update model with provided model_update, where model_update
        should be generated according to `PytorchFLModel.get_model_update()`.

        :param model_update: `ModelUpdate` object that contains the weights \
        that will be used to update the model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if isinstance(model_update, ModelUpdate):
            for p1, p2 in zip(self.get_weights(), model_update.get('weights')):
                p1.data = torch.from_numpy(p2)
                p1.data.requires_grad = True

            if self.use_gpu_for_training and torch.cuda.device_count() > 0:
                self.model.module_.to(self.model.device)
        else:
            raise ValueError('Provided model_update should be of type Model.'
                             'Instead they are:{0}'.format(str(type(model_update))))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        weights = self.get_weights(to_numpy=True)
        update = {'weights': weights}

        return ModelUpdate(**update)

    def predict(self, x):
        """
        Perform prediction for a batch of inputs. Note that for classification
        problems, it returns the resulting probabilities.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`, input compatible with skorch.dataset.Dataset, or pytorch dataloader
        :return: Array of predictions
        :rtype: `np.ndarray`
        """
        if isinstance(x, torch.utils.data.DataLoader):
            return self.predict_generator(x)

        return self.model.predict(x)

    def predict_generator(self, dataloader):
        """
        Performs predictions using a pytorch dataloader

        :param x: A pytorch dataloader with a dataset for predicting
        :type x: 'torch.utils.data.DataLoader'
        :return: Array of predictions
        :rtype: 'np.ndarray'
        """
        predictions = None
        for x_batch, _ in dataloader:
            y_pred = self.model.evaluation_step(x_batch)
            y_pred = y_pred.numpy()
            if predictions is None:
                predictions = y_pred
            else:
                predictions = np.append(predictions, y_pred, 0)

        return predictions

    def evaluate(self, test_dataset, eval_metrics=None, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, y_test) or a pytorch DataLoader
        :type test_dataset: `np.ndarray`
        :param eval_metrics: a list of sklearn.metric class, or a function for evaluating a pytorch dataloader batch
        :type eval_metrics: 'sklearn.metrics' or 'function'
        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test, eval_metrics)

        else:
            return self.evaluate_generator_model(test_dataset, eval_metrics)

    def evaluate_model(self, x, y, eval_metrics=None, **kwargs):
        """
        Evaluates the model given x and y.

        :param x:  Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding labels to x
        :type y: `np.ndarray`
        :param eval_metrics: A list of sklearn.metric class or a functions with the signature \
                        (y_true, y_pred, **kwargs) => float
        :type eval_metrics: `sklearn.metrics` or 'function'
        :return: dictionary of metrics
        :rtype: 'dict'
        """
        if eval_metrics is None or len(eval_metrics) == 0:
            eval_metrics = [metrics.accuracy_score]

        # get names for eval_metrics
        metric_names = [getattr(metric, '__name__', repr(metric))
                        for metric in eval_metrics]

        y_pred = self.predict(x)
        y_pred_exp = np.exp(y_pred)
        y_pred_argmax = np.argmax(y_pred_exp, axis=1)

        try:
            metric_dict = dict(zip(metric_names, eval_metrics))
            # NOTE: Had to replace comprehension with for loop to tackle loss fn case which uses y_pred not argmax
            for metric, fn in metric_dict.items():
                if metric is 'loss':
                    metric_dict[metric] = fn(y, y_pred, self.model)
                else:
                    metric_dict[metric] = fn(y, y_pred_argmax, **kwargs)
            # metric_dict = {metric_name: metric(y, y_pred, **kwargs) for metric_name, metric in
            #                zip(metric_names, eval_metrics)}
            return metric_dict
        except TypeError as exc:
            logger.exception(str(exc))
            raise TypeError("eval_metrics must be an sklearn.eval_metrics class, or a function with the signature "
                            "(y_true, y_pred, **kwargs)'")
        except ValueError as exc:
            logger.exception(exc)
            raise ValueError("arguments not in the correct format for metric")

    def evaluate_generator_model(self, dataloader, eval_metrics=None, **kwargs):
        """
        evaluates the model based on the provided dataloader
        :param dataloader: a pytorch dataloader with test dataset and labels
        :type dataloader: torch.utils.data.DataLoader
        :param eval_metrics: a function for how to evaluate the batched examples and labels
        must have the signature (x_batch, y_batch) => float
        :type eval_metrics: 'function'
        :return: dictionary of metrics
        :rtype: 'dict'
        """
        metric_score = 0
        for x_batch, y_batch in dataloader:
            y_pred = self.model.evaluation_step(x_batch)

            if eval_metrics is None:
                y_pred = np.exp(y_pred)
                y_pred = np.argmax(y_pred, axis=1)
                equals = (y_pred == y_batch)
                metric_score += torch.mean(equals.type(torch.FloatTensor))

            else:
                try:
                    metric_score += eval_metrics(x_batch, y_batch)

                except TypeError as exc:
                    logger.exception(str(exc))
                    raise TypeError("eval_metrics must be a function "
                                    "with the signature (x_batch, y_batch, **kwargs)'")

        metric_dict = {'metric_score': metric_score / len(dataloader)}

        return metric_dict

    def save_model(self, filename=None, optimizer_filename=None, history_filename=None):
        """
        Save a model to file in the format specific to the backend framework.

        :param filename: Name of the file that contains the model to be loaded.
        :type filename: `str`
        :param optimizer_filename: Name of the file that contains the optimizer to be loaded.
        :type optimizer_filename: `str`
        :param history_filename: Name of the file that contains the model history to be loaded.
        :type history_filename: `str`
        :return: None
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            filename = '{}_{}.pt'.format(file, time.time())

        f_params = super().get_model_absolute_path(filename)
        f_optimizer = None
        f_history = None
        if optimizer_filename is not None:
            f_optimizer = super().get_model_absolute_path(optimizer_filename)
        if history_filename is not None:
            f_history = super().get_model_absolute_path(history_filename)
        self.model.save_params(
            f_params=f_params, f_optimizer=f_optimizer, f_history=f_history)
        return filename

    def load_model(self, pytorch_module, model_filename, optimizer_filename=None, history_filename=None, optimizer=None,
                   module_init_params=None):
        """
        Loads a model from disk given the specified file_name

        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference
        :param model_filename: Name of the file that contains the model to be loaded.
        :type model_filename: `str`
        :param optimizer_filename: Name of the file that contains the optimizer to be loaded.
        :type optimizer_filename: `str`
        :param history_filename: Name of the file that contains the model history to be loaded.
        :type history_filename: `str`
        :param optimizer: the optimizer that should be loaded
        :type optimizer: pytorch optimizer class
        :param module_init_params: A dictionary with the values for the model's init arguments. \
        The key for the init parameters must be prefixed with module__ \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :return: None
        """
        f_params = model_filename
        f_optimizer = None
        f_history = None
        if optimizer is not None and optimizer_filename is not None:
            self.optimizer = optimizer
            f_optimizer = optimizer_filename
        if history_filename is not None:
            f_history = history_filename
        model = self.initialize_model(pytorch_module, self.optimizer, self.criterion,
                                      module_init_params=module_init_params)
        model.load_params(f_params=f_params,
                          f_optimizer=f_optimizer, f_history=f_history)

        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            if self.num_gpus > torch.cuda.device_count():
                logger.error('Selected number of gpus to use for training exceeds number of available gpus, ' +
                             str(torch.cuda.device_count()) +
                             'Set number of gpus to maximum available on device.')
                self.num_gpus = torch.cuda.device_count()
            device_ids = list(range(self.num_gpus))
            model.module_ = torch.nn.DataParallel(model.module_, device_ids=device_ids)
            model.set_params(device='cuda')
            model.module_.to('cuda')

        self.model = model

    def get_weights(self, to_numpy=False):
        """
        Returns the weights of the model

        :param to_numpy; Determines whether the weights should be returned as numpy array, or tensor
        :type to_numpy: `boolean`
        :return: list of model weights
        """
        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            module = copy.deepcopy(self.model.module_).cpu()

        else:
            module = self.model.module_
        if to_numpy:
            return self.parameters_to_numpy(module.parameters())
        return list(module.parameters())

    def parameters_to_numpy(self, params):
        """
        Transforms parameter tensors to numpy arrays

        :param params: The parameter tensor to be transformed
        :return: numpy array of parameters 
        """
        np_params = []
        for layer in params:
            np_params.append(layer.detach().numpy())
        return np_params

    def load_model_from_spec(self, model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains one item: model_spec['model_definition'], which has a
        pointer to the file where an nn.sequence container is saved

        :return: model
        :rtype: `nn.sequence`
        """

        model_file = model_spec['model_definition']
        model_absolute_path = config.get_absolute_path(model_file)
        model = torch.load(model_absolute_path)
        return model

    def get_gradient(self, train_data):
        """
        Returns the gradients for each layer in the model

        :return: gradients
        :rtype: `list`. numpy array list of model's gradients
        """

        x = train_data[0]
        y = train_data[1]
        gradients = []
        y_infer = self.model.module_(torch.from_numpy(x))
        loss = self.model.criterion_(y_infer, torch.from_numpy(y))
        loss.backward()

        for layer in self.model.module_.parameters():
            gradients.append(layer.grad.numpy())

        return gradients

    @staticmethod
    def valid_acc(net, x, y):
        """
        Callback scoring function that returns the validation accuracy during training.
        :param net: The model that will be used
        :param x: the validation data set
        :param y: the training targets
        :return: the accuracy of the validation training pass
        """
        y_pred = net.predict(x)
        y_pred = np.exp(y_pred)
        y_pred = np.argmax(y_pred, axis=1)
        return metrics.accuracy_score(y, y_pred)

    @staticmethod
    def loss(y, y_pred, net):
        loss = net.get_loss(torch.from_numpy(y_pred), torch.from_numpy(y)).numpy()

        return loss

    @staticmethod
    def __expand_linear_layer__(net, layer_name, new_dimensions, layer_idx):
        """
        Expands a linear layer of a pytorch module
        :param net: an instantiated pytorch module
        :type net: `torch.nn.Module`
        :param layer_name: layer variable name which needs to be expanded
        :type layer_name: `str`
        :param new_dimensions: new dimensions
        :type new_dimensions: `list`
        :param layer_idx: layer idx which needs to be expanded. this corresponds with
        the index of `new_dimensions`
        :type layer_idx: `int`
        :return: None
        """

        original_lin_layer = getattr(net, layer_name)

        if not isinstance(original_lin_layer, torch.nn.Linear):
            raise FLException('Received a non-linear layer to expand '
                              'whereas the method expects a linear layer')

        new_ip_dim = original_lin_layer.in_features if layer_idx == 0 else new_dimensions[
            layer_idx-1]
        new_op_dim = new_dimensions[layer_idx]
        bias = original_lin_layer.bias is not None

        new_lin_layer = type(original_lin_layer)(
            in_features=new_ip_dim, out_features=new_op_dim, bias=bias
        )

        setattr(net, layer_name, new_lin_layer)

    def expand_model_by_layer_name(self, new_dimension, layer_name="dense"):
        """
        Expands the current PyTorch models layers with the provided dimensions
        :param new_dimension: new dimensions of the particular `layer_name`
        :type new_dimension: `list`
        :param layer_name: layer name which needs to be expanded
        :type layer_name: `str`
        :return: None
        """

        if new_dimension is None:
            raise FLException('No information is provided for '
                              'the new expanded model. '
                              'Please provide the new dimension of '
                              'the resulting expanded model')

        layer_maps = {
            'dense': {
                'class': torch.nn.Linear,
                'expansion_fn': self.__expand_linear_layer__
            }
        }

        layer_cls = layer_maps[layer_name]['class']
        layer_exp_fn = layer_maps[layer_name]['expansion_fn']
        net = self.model.module_    # get instantiated module
        idx = 0

        for layer_varname, layer in net.named_modules():
            # skip layer if not instance of layer desired to be expanded
            if not isinstance(layer, layer_cls):
                continue

            layer_exp_fn(net, layer_varname, new_dimension, idx)
            idx += 1

        self.model.initialize_optimizer()

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not.
        In particular, it calls `skorch.utils.check_is_fitted` to checks
        whether the net is initialized.
        If it has, return True; otherwise return false.

        :return: res
        :rtype: `bool`
        """
        try:
            self.model.check_is_fitted()
        except NotInitializedError:
            return False
        return True

    def get_loss(self, dataset):
        """
        Return the resulting loss computed based on the provided dataset.

        :param train_data: Training data, a tuple \
        given in the form (x_train, y_train). otherwise, input compatible with skorch.dataset.Dataset
        :type train_data: `(np.ndarray, ”np.ndarray)`
        :return: The resulting loss.
        :rtype: `float`
        """
        
        metrics_list = [PytorchFLModel.loss]
        orig_loss_value = self.evaluate(dataset, eval_metrics=metrics_list)[
                    'loss']
        
        return orig_loss_value



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from typing import Union
import sys

from ibm_watson_machine_learning.utils.autoai.utils import try_import_tqdm

try_import_tqdm()

from tqdm import tqdm as TQDM

__all__ = [
    "ProgressBar"
]


class ProgressBar(TQDM):
    """Progress Bar class for handling progress bar display. It is based on 'tqdm' class, could be extended."""

    def __init__(self, ncols: Union[str, int], position: int = 0, desc: str = None, total: int = 100,
                 leave: bool = True, bar_format: str = '{desc}: {percentage:3.0f}%|{bar}|') -> None:
        """
        :param desc: description string to be added as a prefix to progress bar
        :type desc: str, optional

        :param total: total length of the progress bar
        :type total: int, optional
        """
        # note: to see possible progress bar formats please look at super 'bar_format' description
        super().__init__(desc=desc, total=total, leave=leave, position=position, ncols=ncols, file=sys.stdout,
                         bar_format=bar_format)
        self.total = total
        self.previous_message = None
        self.counter = 0
        self.progress = 0

    def increment_counter(self, progress: int = 5) -> None:
        """Increment internal counter and waits for specified time.

        :param progress: number of steps at a time progress bar
        :type progress: int, optional
        """
        self.progress = progress
        self.counter += progress

    def reset_counter(self) -> None:
        """Restart internal counter."""
        self.counter = 0

    def update(self):
        """Update the counter with specific progress."""
        super().update(n=self.progress)

    def last_update(self):
        """Fill up the progress bar till the end, this was the last run."""
        super().update(n=self.total - self.counter)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt
from IPython.display import display, clear_output


def calculate_cum_limits(scores):
    def lower_limit(values):
        mean = np.mean(values)
        std = np.std(values)
        return mean - std

    def upper_limit(values):
        mean = np.mean(values)
        std = np.std(values)
        return mean + std

    n = len(scores) + 1
    lower_limits = [lower_limit(scores[:i]) for i in range(1, n)]
    upper_limits = [upper_limit(scores[:i]) for i in range(1, n)]

    return lower_limits, upper_limits


def plot_subcurve(fig, ax, values, x=None, color='b', title='Learning Curve', xlabel='Training batches', ylabel='Score',
                  refresh=True):
    if x is None:
        x = list(range(0, len(values)))

    lower_limits, upper_limits = calculate_cum_limits(values)
    ax.cla()
    ax.grid()
    ax.plot(x, values, "o-")
    ax.fill_between(
        x,
        lower_limits,
        upper_limits,
        alpha=0.1,
        color=color,
    )
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if refresh:
        display(fig)
        clear_output(wait=True)


def plot_learning_curve(fig, axes, scores, fit_times=None):
    if isinstance(axes, np.ndarray) and fit_times is not None:
        # plot score vs. batch number
        plot_subcurve(fig, axes[0], scores)

        if len(fit_times) < len(scores):
            x = list(range(len(scores) - len(fit_times), len(scores)))
            scores_aligned = scores[len(scores) - len(fit_times):]
        else:
            x = list(range(0, len(scores)))
            scores_aligned = scores

        # plot fit_time vs. batch number
        plot_subcurve(fig, axes[1], fit_times, x, color='g', title='Scalability of the model',
                      xlabel='Training batches', ylabel='Fit time')

        if len(axes) == 3:
            # plot fit_time vs. score
            fit_times_np = np.array(fit_times)
            fit_times_cumsum = np.cumsum(fit_times_np)
            plot_subcurve(fig, axes[2], scores_aligned, x=fit_times_cumsum, color='r', title='Performance of the model',
                          xlabel='Fit time', ylabel='Score')
    else:
        # plot score vs. batch number
        plot_subcurve(fig, axes, scores)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

from pyspark.ml.param import Params
from pyspark.ml.pipeline import Pipeline
from pyspark.ml.wrapper import JavaEstimator
from pyspark.ml.wrapper import _jvm

__all__ = ['Learner', 'Target']

logger = logging.getLogger("mlpipelinepy")

def estimatorCopy(estimator, jvmObject, extra=None):
    if extra is None:
        extra = dict()
    if isinstance(estimator , Pipeline):
        that = Params.copy(estimator, extra)
        stages = [estimatorCopy(stage, jvmObject, extra) for stage in that.getStages()]
        return that.setStages(stages)
    elif (isinstance(estimator,JavaEstimator)):
        newJavaEstimator = estimator.copy(extra)
        paramMapConstructor = jvmObject.org.apache.spark.ml.param.ParamMap
        jParamMap = paramMapConstructor()
        newJavaEstimator._java_obj = newJavaEstimator._java_obj.copy(jParamMap)
        return newJavaEstimator
    else :
        return estimator.copy(extra)


def emptyParam():
    jvmObject = _jvm()
    return jvmObject.org.apache.spark.ml.param.ParamMap.empty()


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

import logging

from ibmfl.exceptions import HyperparamsException

VALID_LOSSES = ["binary_crossentropy", "categorical_crossentropy", "least_squares", "least_absolute_deviation", "auto"]
VALID_SCORES = []

logger = logging.getLogger(__name__)


def validate_parameters(hyperparams):
    """
    Valdiates specific hyperparameters parameters that are defined in
    `hyperparams` variables. Raises an exception for certain parameters that
    are either required or beyond the defined range if they are defined.

    :param hyperparams: Dictionary containing values for the hyperparameters.
    :type  hyperparams: `dict`
    :rtype: None
    """
    try:
        # Check Hyperparameter Type
        if not isinstance(hyperparams, dict):
            raise ValueError("Provided hyperparameter is not valid.")

        # Global Hyperparameter Check
        if "global" in hyperparams:
            params = hyperparams["global"]
        else:
            raise ValueError("Global parameters have not been defined.")

        if "learning_rate" in params:
            if params["learning_rate"] <= 0:
                raise ValueError("learning_rate={} must be strictly " "positive".format(params["learning_rate"]))
        else:
            raise ValueError("learning_rate has not been defined.")

        if "loss" in params:
            if params["loss"] not in VALID_LOSSES:
                raise ValueError(
                    "Loss {} is currently not supported."
                    "Accpted losses: {}".format(params["loss"], ", ".join(VALID_LOSSES))
                )
        else:
            raise ValueError("loss has not been defined.")

        if "num_classes" in params:
            # Validate Class Parameter Types
            if type(params["num_classes"]) is not int:
                raise ValueError(
                    "Provided classes value's type is not valid, " "should be an int value >= 2 for classification."
                )

            # Validate Classification Case
            if params["loss"] != "least_squares":
                if params["num_classes"] < 0:
                    raise ValueError("Provided class value must be >= 2 for " "classification.")

                if params["loss"] == "binary_crossentropy" and params["num_classes"] != 2:
                    raise ValueError("Binary class models must have class of 2.")
                elif params["loss"] == "categorical_crossentropy" and params["num_classes"] <= 2:
                    raise ValueError("Multiclass models must have class > 2.")
                elif params["loss"] == "auto":
                    if params["num_classes"] < 2:
                        raise ValueError("Class value must be >= 2.")
                    else:
                        logging.warning(
                            "Obtaining class labels based on local dataset. "
                            "This may cause failures during aggregation "
                            "when parties have distinctive class labels."
                        )
        else:
            # Handle Classes Not Defined Case
            if params["loss"] != "least_squares":
                raise ValueError(
                    "Classes has not been defined. Should provide " "a value >= 2 for classification models."
                )

        if "max_bins" in params:
            if not (2 <= params["max_bins"] and params["max_bins"] <= 255):
                raise ValueError(
                    "max_bins={} should be no smaller than 2 " "and no larger than 255.".format(params["max_bins"])
                )

        if "max_iter" in params:
            if params["max_iter"] < 1:
                raise ValueError("max_iter={} must not be smaller " "than 1.".format(params["max_iter"]))
        else:
            raise ValueError("max_iter has not been defined.")

        if "max_depth" in params:
            if params["max_depth"] is not None and params["max_depth"] < 0:
                raise ValueError("max_depth={} must be strictly greater" "than 0.".format(params["max_depth"]))

        if "max_leaf_nodes" in params:
            if params["max_leaf_nodes"] is not None and params["max_leaf_nodes"] <= 1:
                raise ValueError(
                    "max_leaf_nodes={} must be strictly greater" "than 1.".format(params["max_leaf_nodes"])
                )

        if "min_samples_leaf" in params:
            if params["min_samples_leaf"] is not None and params["min_samples_leaf"] < 0:
                raise ValueError("min_sample_leaf={} must not be smaller " "than 0".format(params["min_samples_leaf"]))

        if "data_sketch_accuracy" in params and params["data_sketch_accuracy"] is not None:
            if params["data_sketch_accuracy"] <= 0 or params["data_sketch_accuracy"] >= 1:
                raise ValueError("data_sketch_accuracy={} must be strictly greater than 0 and less than 1")

    except Exception as ex:
        logger.exception(str(ex))
        raise HyperparamsException("Defined global hyperparameters malformed. " + str(ex))


def init_parameters(obj, hyperparameters):
    """
    Initializes the hyperparameters to the corresponding object as an attribute.
    The parameters are propagated by reference, so changes should propagate to
    respective object passed into the function.

    :param obj: Any class object which to tie initialize the hyperparameters with.
    :type  obj: Object
    :param hyperparameters: Dictionary containing values for the hyperparameters.
    :type  hyperparameters: `dict`
    """
    # Initialize Global Configuration Parameter
    params = hyperparameters["global"]
    setattr(obj, "param", params)

    # Initialize Attributes (Pre-Checked Parameters)
    setattr(obj, "learning_rate", params["learning_rate"])
    setattr(obj, "loss", params["loss"])
    setattr(obj, "max_iter", params["max_iter"])

    if params["loss"] == "least_squares":
        setattr(obj, "num_classes", 1)
    elif params["loss"] in ["binary_crossentropy", "categorical_crossentropy", "auto"]:
        setattr(obj, "num_classes", params["num_classes"])

    # Initialize Attributes (Optional Values - Based on Default Parameters)
    if "l2_regularization" not in params or params["l2_regularization"] is None:
        setattr(obj, "l2_regularization", 0)
    else:
        setattr(obj, "l2_regularization", params["l2_regularization"])

    if "max_bins" not in params:
        setattr(obj, "max_bins", 255)
    else:
        setattr(obj, "max_bins", params["max_bins"])

    if "max_depth" not in params or params["max_depth"] is None:
        setattr(obj, "max_depth", None)
    else:
        setattr(obj, "max_depth", params["max_depth"])

    if "max_leaf_nodes" not in params or params["max_leaf_nodes"] is None:
        setattr(obj, "max_leaf_nodes", 31)
    else:
        setattr(obj, "max_leaf_nodes", params["max_leaf_nodes"])

    if "min_samples_leaf" not in params or params["min_samples_leaf"] is None:
        setattr(obj, "min_samples_leaf", 20)
    else:
        setattr(obj, "min_samples_leaf", params["min_samples_leaf"])

    if "random_state" in params:
        setattr(obj, "random_state", params["random_state"])
    else:
        setattr(obj, "random_state", None)

    if "scoring" in params:
        setattr(obj, "scoring", params["scoring"])
    else:
        setattr(obj, "scoring", None)

    if "verbose" not in params or params["verbose"] is None:
        setattr(obj, "verbose", False)
    else:
        setattr(obj, "verbose", True)

    if "data_sketch_accuracy" not in params or params["data_sketch_accuracy"] is None:
        setattr(obj, "data_sketch_accuracy", None)  # default
    else:
        setattr(obj, "data_sketch_accuracy", params["data_sketch_accuracy"])

    return obj



"""
Pickle based serialization
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import pickle
from ibmfl.message.message import Message
from ibmfl.message.serializer import Serializer


class PickleSerializer(Serializer):
    """
    Class for Pickle based serialization
    """

    def serialize(self, msg):
        """
        Serialize a message using pickle

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialize byte stream
        :rtype: `b[]`
        """
        msg_header = msg.get_header()
        serialized_data = msg.get_data()  # need to serialize the data
        return pickle.dumps({'header': msg_header,
                             'data': serialized_data,
                             })

    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        data_dict = pickle.loads(serialized_byte_stream)
        if 'MSG_LEN|' in data_dict:
            msg_length = int(data_dict.split('|')[1])
            return msg_length

        msg = Message(data=data_dict['data'])
        msg.set_header(data_dict['header'])

        return msg



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

from pyspark import SparkContext, SQLContext
from pyspark.sql import DataFrame
from pyspark.sql.types import _parse_datatype_json_string

__all__ = ['DataEdge', "MetaEdge"]

logger = logging.getLogger("mlpipelinepy")


class DataEdge(object):
    """
    DataEdge information
    """

    def __init__(self, s_tuple_id_df):
        try:
            self._java_obj = s_tuple_id_df
            self._id = s_tuple_id_df._1()
            self.j_Data_Frame = s_tuple_id_df._2()
        except Exception as e:
            logger.exception(e)
            raise

    @property
    def id(self):
        return self._id

    @property
    def data_frame(self):
        sc = SparkContext.getOrCreate()
        ctx = SQLContext.getOrCreate(sc)
        return DataFrame(self.j_Data_Frame, ctx)


class MetaEdge(object):
    """
    MetaEdge information
    """

    def __init__(self, s_meta_edge):
        try:
            self._id = str(s_meta_edge._1())
            self._fromNode = None
            self._schema = _parse_datatype_json_string(s_meta_edge._2().json())
        except Exception as e:
            logger.exception(e)
            raise

    @property
    def id(self):
        return self._id

    @property
    def schema(self):
        return self._schema

    def __str__(self):
        return "Id: " + self._id + " Schema: " + str(self._schema)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from setuptools import setup, find_packages
import os

with open('VERSION', 'r') as f_ver:
    VERSION = f_ver.read()


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as  f:
        return f.read()


# read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

extras_require = {
    "fl": [
        "tensorflow==2.4.4",
        "scikit-learn==0.23.2",
        "torch==1.7.1",
        "numpy==1.19.2",
        "pandas==1.3.4",
        "pytest==6.2.5",
        "pyYAML==5.4.1",
        "parse==1.19.0",
        "websockets==8.1",
        "jsonpickle==1.4.2",
        "requests==2.27.1",
        "scipy==1.6.3",
        "environs==9.5.0",
        "pathlib2==2.3.6",
        "diffprivlib==0.5.1",
        "numcompress==0.1.2",
        "psutil",
        "setproctitle",
        "tabulate==0.8.9",
        "lz4",
        "gym",
        "cloudpickle==1.3.0",
        "image==1.5.33",
        "ddsketch==1.1.2",
        "skorch==0.11.0",
    ],

    "fl-rt22.1": [
        "tensorflow==2.7.2",
        "scikit-learn==1.0.2",
        "torch==1.10.2",
        "numpy==1.20.3",
        "pandas==1.3.4",
        "pytest==6.2.5",
        "pyYAML==5.4.1",
        "parse==1.19.0",
        "websockets==10.1",
        "jsonpickle==1.4.2",
        "requests==2.27.1",
        "scipy==1.7.3",
        "environs==9.5.0",
        "pathlib2==2.3.6",
        "diffprivlib==0.5.1",
        "numcompress==0.1.2",
        "psutil",
        "setproctitle",
        "tabulate==0.8.9",
        "lz4",
        "gym",
        "cloudpickle==1.3.0",
        "image==1.5.33",
        "ddsketch==1.1.2",
        "skorch==0.11.0",
        "protobuf==3.19.5",
        "GPUtil",
        "joblib==1.1.1",
    ],

    "fl-rt22.2-py3.10": [
        "tensorflow==2.9.3",
        "scikit-learn==1.1.1",
        "torch==1.12.1",
        "numpy==1.23.1",
        "pandas==1.4.3",
        "pytest==6.2.5",
        "pyYAML==5.4.1",
        "parse==1.19.0",
        "websockets==10.1",
        "jsonpickle==1.4.2",
        "requests==2.27.1",
        "scipy==1.8.1",
        "environs==9.5.0",
        "pathlib2==2.3.6",
        "diffprivlib==0.5.1",
        "numcompress==0.1.2",
        "psutil",
        "setproctitle",
        "tabulate==0.8.9",
        "lz4",
        "gym",
        "cloudpickle==1.3.0",
        "image==1.5.33",
        "ddsketch==1.1.2",
        "skorch==0.12.0",
        "protobuf==3.19.5",
        "GPUtil",
        "joblib==1.1.1",
        "cryptography==39.0.1"
    ],

    "fl-rt23.1-py3.10": [
        "tensorflow==2.12.0",
        "scikit-learn==1.1.1",
        "torch==2.0.0",
        "numpy==1.23.5",
        "pandas==1.5.3",
        "pytest==6.2.5",
        "pyYAML==5.4.1",
        "parse==1.19.0",
        "websockets==10.1",
        "jsonpickle==1.4.2",
        "requests==2.27.1",
        "scipy==1.10.1",
        "environs==9.5.0",
        "pathlib2==2.3.6",
        "diffprivlib==0.5.1",
        "numcompress==0.1.2",
        "psutil",
        "setproctitle",
        "tabulate==0.8.9",
        "lz4",
        "gym",
        "cloudpickle==1.3.0",
        "image==1.5.33",
        "ddsketch==1.1.2",
        "skorch==0.12.0",
        "protobuf==4.22.1",
        "GPUtil",
        "joblib==1.1.1",
        "cryptography==39.0.1"
    ],

    "fl-crypto": [
        "pyhelayers==1.5.0.3"
    ],
}


def retrieve_all_files_from_path_and_sub_paths(path):
    listOfFiles = list()
    for (dirpath, dirnames, filenames) in os.walk(path):
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]

    return listOfFiles


with open('MANIFEST.in', 'a') as f:
    f.writelines([f"exclude {line}\n" for line in
                  retrieve_all_files_from_path_and_sub_paths('ibm_watson_machine_learning/tests')])

setup(
    name='ibm_watson_machine_learning',
    version=VERSION,
    python_requires='>=3.9',
    author='IBM',
    author_email='lukasz.cmielowski@pl.ibm.com, dorota.laczak@ibm.com',
    description='IBM Watson Machine Learning API Client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Natural Language :: English',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Microsoft :: Windows',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Information Technology',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Internet'],
    keywords=["watson", "machine learning", "IBM", "Bluemix", "client", "API", "IBM Cloud"],
    url='https://ibm.github.io/watson-machine-learning-sdk',
    packages=find_packages(),
    package_data={'': ['messages/messages_en.json']},
    install_requires=[
        'requests',
        'urllib3',
        'pandas<1.6.0,>=0.24.2',
        'certifi',
        'lomond',
        'tabulate',
        'packaging',
        "ibm-cos-sdk<2.14.0,>=2.12.0; python_version>='3.10'",
        "ibm-cos-sdk==2.11.*; python_version>='3.9' and python_version<'3.10'",
        "ibm-cos-sdk==2.7.*; python_version<'3.9'",
        'importlib-metadata'
    ],
    include_package_data=True,
    extras_require=extras_require,
)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2019- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import PIPELINE_DETAILS_TYPE
from ibm_watson_machine_learning.metanames import PipelineMetanames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.href_definitions import API_VERSION, SPACES, PIPELINES, LIBRARIES, EXPERIMENTS, RUNTIMES, DEPLOYMENTS
from ibm_watson_machine_learning.wml_client_error import WMLClientError
_DEFAULT_LIST_LENGTH = 50
import json


class Pipelines(WMLResource):
    """Store and manage pipelines."""

    ConfigurationMetaNames = PipelineMetanames()
    """MetaNames for pipelines creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)

        self.WSD = client.WSD
        self._ICP = client.ICP

    def _generate_pipeline_document(self, meta_props):
        if self._client.ICP:
            doc = {
                "doc_type": "pipeline",
                "version": "2.0",
                "primary_pipeline": "wmla_only",
                "pipelines": [
                    {
                        "id": "wmla_only",
                        "runtime_ref": "hybrid",
                        "nodes": [
                            {
                                "id": "training",
                                "type": "model_node",
                                "op": "dl_train",
                                "runtime_ref": "DL_WMLA",
                                "inputs": [
                                ],
                                "outputs": [],
                                "parameters": {
                                    "name": "pipeline",
                                    "description": "Pipeline - Python client"
                                }
                            }
                        ]
                    }
                ],
                "schemas": [
                    {
                        "id": "schema1",
                        "fields": [
                            {
                                "name": "text",
                                "type": "string"
                            }
                        ]
                    }
                ]
            }
        else:
            doc = {
                "doc_type": "pipeline",
                "version": "2.0",
                "primary_pipeline": "dlaas_only",
                "pipelines": [
                  {
                    "id": "dlaas_only",
                    "runtime_ref": "hybrid",
                    "nodes": [
                      {
                        "id": "training",
                        "type": "model_node",
                        "op": "dl_train",
                        "runtime_ref": "DL",
                        "inputs": [
                        ],
                        "outputs": [],
                        "parameters": {
                          "name": "tf-mnist",
                          "description": "Simple MNIST model implemented in TF"
                        }
                      }
                    ]
                  }
                ],
                "schemas": [
                  {
                    "id": "schema1",
                    "fields": [
                      {
                        "name": "text",
                        "type": "string"
                      }
                    ]
                  }
                ]
              }
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_35 or self._client.ICP_40 or \
                self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47:
            if self.ConfigurationMetaNames.COMMAND in meta_props:
                doc["pipelines"][0]["nodes"][0]["parameters"]["command"] = meta_props[self.ConfigurationMetaNames.COMMAND]
            if self.ConfigurationMetaNames.RUNTIMES in meta_props:
                doc["runtimes"] = meta_props[self.ConfigurationMetaNames.RUNTIMES]
                if self._client.ICP:
                    doc["runtimes"][0]["id"] = "DL_WMLA"
                else:
                    doc["runtimes"][0]["id"] = "DL"
            if self.ConfigurationMetaNames.LIBRARY_UID in meta_props:
                if self._client.CAMS:
                    type_uid = self._check_if_lib_or_def(meta_props[self.ConfigurationMetaNames.LIBRARY_UID])
                    doc["pipelines"][0]["nodes"][0]["parameters"]["training_lib_href"] = type_uid
                else:
                    doc["pipelines"][0]["nodes"][0]["parameters"]["training_lib_href"] = "/v4/libraries/" + meta_props[
                        self.ConfigurationMetaNames.LIBRARY_UID]
            if self.ConfigurationMetaNames.COMPUTE in meta_props:
                doc["pipelines"][0]["nodes"][0]["parameters"]["compute"] = meta_props[self.ConfigurationMetaNames.COMPUTE]
            return doc
        else:
            return doc

    def store(self, meta_props, **kwargs):
        """Create a pipeline.

        :param meta_props: meta data of the pipeline configuration. To see available meta names use:

            .. code-block:: python

                client.pipelines.ConfigurationMetaNames.get()

        :type meta_props: dict

        :return: stored pipeline metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.pipelines.ConfigurationMetaNames.NAME: 'my_training_definition',
                client.pipelines.ConfigurationMetaNames.DOCUMENT: {"doc_type":"pipeline",
                                                                   "version": "2.0",
                                                                   "primary_pipeline": "dlaas_only",
                                                                   "pipelines": [{"id": "dlaas_only",
                                                                                  "runtime_ref": "hybrid",
                                                                                  "nodes": [{"id": "training",
                                                                                             "type": "model_node",
                                                                                             "op": "dl_train",
                                                                                             "runtime_ref": "DL",
                                                                                             "inputs": [],
                                                                                             "outputs": [],
                                                                                             "parameters": {"name": "tf-mnist",
                                                                                                            "description": "Simple MNIST model implemented in TF",
                                                                                                            "command": "python3 convolutional_network.py --trainImagesFile ${DATA_DIR}/train-images-idx3-ubyte.gz --trainLabelsFile ${DATA_DIR}/train-labels-idx1-ubyte.gz --testImagesFile ${DATA_DIR}/t10k-images-idx3-ubyte.gz --testLabelsFile ${DATA_DIR}/t10k-labels-idx1-ubyte.gz --learningRate 0.001 --trainingIters 6000",
                                                                                                            "compute": {"name": "k80","nodes": 1},
                                                                                                            "training_lib_href": "/v4/libraries/64758251-bt01-4aa5-a7ay-72639e2ff4d2/content"
                                                                                             },
                                                                                             "target_bucket": "wml-dev-results"
                                                                                  }]
                                                                   }]
                }
            }
            pipeline_details = client.pipelines.store(training_definition_filepath, meta_props=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        Pipelines._validate_type(meta_props, u'meta_props', dict, True)

        if self.ConfigurationMetaNames.DOCUMENT in meta_props:
            pipeline_meta = self.ConfigurationMetaNames._generate_resource_metadata(
                meta_props,
                with_validation=True,
                client=self._client

            )
        else:
            document = self._generate_pipeline_document(meta_props)
            meta_props[self.ConfigurationMetaNames.DOCUMENT] = document
            pipeline_meta = self.ConfigurationMetaNames._generate_resource_metadata(
                meta_props,
                with_validation=True,
                client=self._client
            )
        if self._client.CAMS:
            if self._client.default_space_id is not None:
                pipeline_meta['space'] = {'href': "/v4/spaces/"+self._client.default_space_id}
            elif self._client.default_project_id is not None:
                pipeline_meta['project'] = {'href': "/v2/projects/"+self._client.default_project_id}
            else:
                raise WMLClientError("It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_35 or self._client.ICP_40 or \
                self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47:
            if self._client.default_space_id is not None:
                pipeline_meta['space_id'] = self._client.default_space_id
            elif self._client.default_project_id is not None:
                pipeline_meta['project_id'] = self._client.default_project_id
            else:
                raise WMLClientError(
                    "It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")

        # add kwargs into optimization section at the very end of preparing payload
        try:
            for p in pipeline_meta[self.ConfigurationMetaNames.DOCUMENT]['pipelines']:
                for n in p['nodes']:
                    params = n['parameters']['optimization']
                    params.update(kwargs)
                    n['parameters']['optimization'] = params
        except:
            pass

        if not self._ICP and not self._client.WSD:
            if self._client.CLOUD_PLATFORM_SPACES:
                creation_response = requests.post(
                    self._client.service_instance._href_definitions.get_pipelines_href(),
                    headers=self._client._get_headers(),
                    params=self._client._params(skip_for_create=True),
                    json=pipeline_meta
                 )
            else:
                creation_response = requests.post(
                    self._wml_credentials['url'] + '/v4/pipelines',
                    headers=self._client._get_headers(),
                    json=pipeline_meta
                )

        else:
            if self._client.WSD:
                if "space_uid" in meta_props:
                    raise WMLClientError(
                        u'Invalid input SPACE_UID in meta_props. SPACE_UID not supported for Watson Studio Desktop.')

                pipeline_url = self._wml_credentials['url'] + '/v4/pipelines'
                if self._client.default_project_id is not None:
                    pipeline_meta['project'] = {'href': "/v2/projects/" + self._client.default_project_id}
                else:
                    raise WMLClientError(
                        "It is mandatory to set the project id. Use client.set.default_project(<PROJECT_UID>) to proceed.")

                import copy
                cams_entity = copy.deepcopy(pipeline_meta)
                pipeline_url = self._client.service_instance._href_definitions.get_wsd_model_href()
                #cams_entity.pop('name')
                if cams_entity.get('description') is not None:
                    cams_entity.pop('description')
                cams_payload = self._create_wsd_payload('wml_pipeline', cams_entity, pipeline_meta)

                atype_body = {
                    "name": "wml_pipeline"
                }
                atype_payload = json.dumps(atype_body, separators=(',', ':'))
                atype_url = self._client.service_instance._href_definitions.get_wsd_asset_type_href()
                aheaders = {
                    'Content-Type': "application/json"
                }
                asset_type_response = requests.post(
                    atype_url,
                    params=self._client._params(),
                    data=atype_payload,
                    headers=aheaders
                )

                if asset_type_response.status_code != 200 and \
                        asset_type_response.status_code != 201 and asset_type_response.status_code != 409:
                    raise WMLClientError("Failed to create asset type. Try again.")

                creation_response = requests.post(
                    pipeline_url,
                    params=self._client._params(),
                    headers=self._client._get_headers(),
                    json=cams_payload
                )
                pipeline_details = self._handle_response(201, u'creating new pipeline', creation_response)
                return WMLResource._wsd_get_required_element_from_response(self, pipeline_details)

            elif self._client.ICP_35 or self._client.ICP_40 or self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47:
                creation_response = requests.post(
                    self._client.service_instance._href_definitions.get_pipelines_href(),
                    headers=self._client._get_headers(),
                    params=self._client._params(skip_for_create=True),
                    json=pipeline_meta
                 )
            else:
                creation_response = requests.post(
                    self._wml_credentials['url'] + '/v4/pipelines',
                    headers=self._client._get_headers(),
                    json=pipeline_meta
                )
        pipeline_details = self._handle_response(201, u'creating new pipeline', creation_response)

        return pipeline_details

    def create_revision(self, pipeline_uid):
        """Create a new pipeline revision.

        :param pipeline_uid: Unique pipeline ID
        :type pipeline_uid: str

        :return: pipeline revision details
        :rtype: dict

        **Example**

        .. code-block:: python

            client.pipelines.create_revision(pipeline_uid)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        Pipelines._validate_type(pipeline_uid, u'pipeline_uid', str, False)

        if not any([self._client.CLOUD_PLATFORM_SPACES, self._client.ICP_35, self._client.ICP_40,
                    self._client.ICP_45, self._client.ICP_46, self._client.ICP_47]):
            raise WMLClientError(
                u'Revision support is not there in this version WML server. It is supported only from 3.1.0 onwards.')
        else:
            url = self._client.service_instance._href_definitions.get_pipelines_href()
            #return self. _create_revision_artifact_for_assets(pipeline_uid, 'pipelines')
            return self._create_revision_artifact(url, pipeline_uid, 'pipelines')

    def update(self, pipeline_uid, changes, rev_id = None):
        """Updates existing pipeline metadata.

        :param pipeline_uid: Unique Id of pipeline which definition should be updated
        :type pipeline_uid: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated pipeline
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.pipelines.ConfigurationMetaNames.NAME: "updated_pipeline"
            }
            pipeline_details = client.pipelines.update(pipeline_uid, changes=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        if self._client.WSD:
            raise WMLClientError(" Update pipeline operation is not supported in WSD. ")

        self._validate_type(pipeline_uid, u'pipeline_uid', str, True)
        self._validate_type(changes, u'changes', dict, True)

        details = self._client.pipelines.get_details(pipeline_uid)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes,
                                                                            with_validation=True)

        url = self._client.service_instance._href_definitions.get_pipeline_href(pipeline_uid)

        response = requests.patch(url, json=patch_payload, params=self._client._params(),headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'pipeline patch', response)

        return updated_details

    def delete(self, pipeline_uid):
        """Delete a stored pipeline.

        :param pipeline_uid: Unique Id of pipeline
        :type pipeline_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.pipelines.delete(pipeline_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        Pipelines._validate_type(pipeline_uid, u'pipeline_uid', str, True)
        if self._client.WSD:
            pipeline_endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + pipeline_uid
        else:
            pipeline_endpoint = self._client.service_instance._href_definitions.get_pipeline_href(pipeline_uid)

        response_delete = requests.delete(pipeline_endpoint, params=self._client._params(), headers=self._client._get_headers())

        return self._handle_response(204, u'pipeline deletion', response_delete, False)

    def get_details(self, pipeline_uid=None, limit=None, asynchronous=False, get_all=False):
        """Get metadata of stored pipeline(s). If pipeline UID is not specified returns all pipelines metadata.

        :param pipeline_uid: Pipeline UID
        :type pipeline_uid: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional
        :param asynchronous: if `True`, it will work as a generator
        :type asynchronous: bool, optional
        :param get_all: if `True`, it will get all entries in 'limited' chunks
        :type get_all: bool, optional

        :return: metadata of pipeline(s)
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)

        **Example**

        .. code-block:: python

            pipeline_details = client.pipelines.get_details(pipeline_uid)
            pipeline_details = client.pipelines.get_details()
            pipeline_details = client.pipelines.get_details(limit=100)
            pipeline_details = client.pipelines.get_details(limit=100, get_all=True)
            pipeline_details = []
            for entry in client.pipelines.get_details(limit=100, asynchronous=True, get_all=True):
                pipeline_details.extend(entry)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        Pipelines._validate_type(pipeline_uid, u'pipeline_uid', str, False)
        Pipelines._validate_type(limit, u'limit', int, False)
        if self._client.WSD:
            url = self._client.service_instance._href_definitions.get_model_definition_assets_href()
            response = self._get_artifact_details(url, pipeline_uid, limit, 'pipeline' if pipeline_uid else 'pipelines', summary="False")
            return self._wsd_get_required_element_from_response(response)
        else:
            url = self._client.service_instance._href_definitions.get_pipelines_href()

        if pipeline_uid is None:
            return self._get_artifact_details(url, pipeline_uid, limit, 'pipelines', summary="False",
                                              _async=asynchronous, _all=get_all)
        else:
            return self._get_artifact_details(url, pipeline_uid, limit, 'pipeline',summary="False")

    def get_revision_details(self, pipeline_uid, rev_uid):
        """Get metadata of pipeline revision.

        :param pipeline_uid:  stored pipeline UID
        :type pipeline_uid: str

        :param rev_uid:  stored pipeline revision ID
        :type rev_uid: str

        :return: stored pipeline revision metadata
        :rtype: dict

        **Example:**

        .. code-block:: python

            pipeline_details = client.pipelines.get_revision_details(pipeline_uid, rev_uid)

        .. note::
            `rev_uid` parameter is not applicable in Cloud platform.
         """
        Pipelines._validate_type(pipeline_uid, u'pipeline_uid', str, True)
        Pipelines._validate_type(rev_uid, u'rev_uid', int, True)
        if not any([self._client.CLOUD_PLATFORM_SPACES, self._client.ICP_35, self._client.ICP_40,
                    self._client.ICP_45, self._client.ICP_46, self._client.ICP_47]):
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        else:
            url = self._client.service_instance._href_definitions.get_pipeline_href(pipeline_uid)
        return self._get_with_or_without_limit(url, limit=None, op_name="pipeline",
                                               summary=None, pre_defined=None, revision=rev_uid)

    @staticmethod
    def get_href(pipeline_details):
        """Get href from pipeline details.

        :param pipeline_details: metadata of the stored pipeline
        :type pipeline_details: dict

        :return: pipeline href
        :rtype: str

        **Example**

        .. code-block:: python

            pipeline_details = client.pipelines.get_details(pipeline_uid)
            pipeline_href = client.pipelines.get_href(pipeline_details)

        """
        Pipelines._validate_type(pipeline_details, u'pipeline_details', object, True)

        if 'asset_type' in pipeline_details['metadata']:
            return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details',
                                                               [u'metadata', u'href'])
        else:
            if 'href' in pipeline_details['metadata']:
                Pipelines._validate_type_of_details(pipeline_details, PIPELINE_DETAILS_TYPE)
                return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details', [u'metadata', u'href'])
            else:
                pipeline_id = WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details', [u'metadata', u'id'])
                return "/ml/v4/pipelines/" + pipeline_id


    @staticmethod
    def get_uid(pipeline_details):
        """Get pipeline_uid from pipeline details.

        :param pipeline_details: metadata of the stored pipeline
        :type pipeline_details: dict

        :return: Unique Id of pipeline
        :rtype: str

        **Example**

        .. code-block:: python

            pipeline_uid = client.pipelines.get_uid(pipeline_details)

        """
        Pipelines._validate_type(pipeline_details, u'pipeline_details', object, True)
        if 'asset_id' in pipeline_details['metadata']:
            return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details',
                                                               [u'metadata', u'asset_id'])
        else:
            if 'id' not in pipeline_details[u'metadata']:
                Pipelines._validate_type_of_details(pipeline_details, PIPELINE_DETAILS_TYPE)

            return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details', [u'metadata', u'id'])

    @staticmethod
    def get_id(pipeline_details):
        """Get pipeline id from pipeline details.

        :param pipeline_details: metadata of the stored pipeline
        :type pipeline_details: dict

        :return: Unique Id of pipeline
        :rtype: str

        **Example**

        .. code-block:: python

            pipeline_uid = client.pipelines.get_id(pipeline_details)

        """
        Pipelines._validate_type(pipeline_details, u'pipeline_details', object, True)
        if 'asset_id' in pipeline_details['metadata']:
            return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details',
                                                               [u'metadata', u'asset_id'])
        else:
            if 'id' not in pipeline_details[u'metadata']:
                Pipelines._validate_type_of_details(pipeline_details, PIPELINE_DETAILS_TYPE)

            return WMLResource._get_required_element_from_dict(pipeline_details, u'pipeline_details',
                                                               [u'metadata', u'id'])

    def list(self, limit=None, return_as_df=True):
        """Print stored pipelines in a table format. If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed pipelines or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.pipelines.list()

        """
        table = None
        if self._client.WSD:
            Pipelines._wsd_list_assets(self, "wml_pipeline", limit)
        else:
            ##For CP4D, check if either spce or project ID is set
            self._client._check_if_either_is_set()

            pipeline_resources = self.get_details(limit=limit)[u'resources']
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_35 or self._client.ICP_40 or \
                    self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47:
                pipeline_values = [(m[u'metadata'][u'id'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at']) for m
                                   in pipeline_resources]

                table = self._list(pipeline_values, [u'ID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)
            else:
                pipeline_values = [(m[u'metadata'][u'id'], m[u'entity'][u'name'], m[u'metadata'][u'created_at']) for m in pipeline_resources]

                table = self._list(pipeline_values, [u'GUID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    def list_revisions(self, pipeline_uid, limit=None, return_as_df=True):
        """Print all revision for the given pipeline uid in a table format.

        :param pipeline_uid: Unique id of stored pipeline
        :type pipeline_uid: str

        :param limit: limit number of fetched records
        :type limit: int, optional

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed revisions or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.pipelines.list_revisions(pipeline_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        Pipelines._validate_type(pipeline_uid, u'pipeline_uid', str, True)
        table = None
        if not any([self._client.CLOUD_PLATFORM_SPACES, self._client.ICP_35, self._client.ICP_40,
                    self._client.ICP_45, self._client.ICP_46, self._client.ICP_47]):
            raise WMLClientError(
                u'Revision support is not there in this WML server. It is supported only from 3.1.0 onwards.')
        else:
            url = self._client.service_instance._href_definitions.get_pipeline_href(pipeline_uid)
            pipeline_resources = self._get_artifact_details(url, "revisions", limit, 'pipeline revisions')[u'resources']
            pipeline_values = [
                (m[u'metadata'][u'rev'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at']) for m in
                pipeline_resources]

            table = self._list(pipeline_values, [u'GUID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    def clone(self, pipeline_uid, space_id=None, action="copy", rev_id=None):
        """Create a new pipeline identical with the given pipeline either in the same space or in a new space.
        All dependent assets will be cloned too.

        :param pipeline_uid: uid of the pipeline to be cloned
        :type pipeline_uid: str
        :param space_id: uid of the space to which the pipeline needs to be cloned
        :type space_id: str, optional
        :param action: action specifying "copy" or "move"
        :type action: str, optional
        :param rev_id: revision ID of the pipeline
        :type rev_id: str, optional

        :return: metadata of the pipeline cloned
        :rtype: dict

        **Example**

        .. code-block:: python

            client.pipelines.clone(pipeline_uid=artifact_id,space_id=space_uid,action="copy")

        .. note::
            * If revision id is not specified, all revisions of the artifact are cloned.

            * Space id is mandatory for "move" action.

        """
        if any([self._client.WSD, self._client.CLOUD_PLATFORM_SPACES, self._client.ICP_35, self._client.ICP_40,
                self._client.ICP_45, self._client.ICP_46, self._client.ICP_47]):
            raise WMLClientError(" Clone pipeline operation is not supported. ")
        Pipelines._validate_type(pipeline_uid, 'exp_id', str, True)
        clone_meta = {}
        if space_id is not None:
            clone_meta["space"] = {"href": API_VERSION + SPACES + "/" + space_id}
        if action is not None:
            clone_meta["action"] = action
        if rev_id is not None:
            clone_meta["rev"] = rev_id

        url = self._client.service_instance._href_definitions.get_pipeline_href(pipeline_uid)

        response_post = requests.post(url, json=clone_meta,
                                          headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=200, operationName=u'cloning pipeline',
                                            response=response_post)

        return details